# Generated by Django 3.1.7 on 2021-08-06 08:51

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0155_merge_20210728_1430'),
    ]

    operations = [
        migrations.AlterField(
            model_name='opportunity',
            name='name',
            field=models.CharField(db_index=True, max_length=100),
        ),
        migrations.AlterField(
            model_name='ticket',
            name='summary',
            field=models.CharField(blank=True, db_index=True, max_length=250, null=True),
        ),
    ]



# Generated by Django 4.0.4 on 2022-08-08 07:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0166_rename_initial_ticket_initial_description_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='ticket',
            name='is_issue_flag',
            field=models.BooleanField(default=False),
        ),
    ]



# Generated by Django 2.1 on 2019-07-19 10:58

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0101_agreement_company'),
    ]

    operations = [
        migrations.AddField(
            model_name='agreement',
            name='agreement_type',
            field=models.CharField(max_length=50, null=True),
        ),
        migrations.AddField(
            model_name='agreement',
            name='cancelled_flag',
            field=models.BooleanField(default=False),
        ),
    ]



# Generated by Django 2.0 on 2018-08-23 10:25

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0071_auto_20180822_1300'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='callbackentry',
            name='member',
        ),
    ]



import datetime
import json
import logging
import re
from urllib.parse import urlparse
from json import JSONDecodeError


import pytz
import requests
from django.conf import settings
from django.core.cache import cache
from django.db import models
from retrying import retry

from djconnectwise.utils import DjconnectwiseSettings

# Cloud URLs:
# https://developer.connectwise.com/Products/Manage/Developer_Guide#Cloud_URLs
CW_CLOUD_URLS = {
    'au.myconnectwise.net': 'api-au.myconnectwise.net',
    'eu.myconnectwise.net': 'api-eu.myconnectwise.net',
    'na.myconnectwise.net': 'api-na.myconnectwise.net',
    'aus.myconnectwise.net': 'api-aus.myconnectwise.net',
    'za.myconnectwise.net': 'api-za.myconnectwise.net',
}
COMPANY_INFO_REQUIRED = 'company-info-required'
CW_CLOUD_DOMAIN = 'myconnectwise.net'
DEFAULT_CW_API_CODEBASE = 'v4_6_release/'

CW_RESPONSE_MAX_RECORDS = 1000  # The greatest number of records ConnectWise
# will send us in one response.
RETRY_WAIT_EXPONENTIAL_MULTAPPLIER = 1000  # Initial number of milliseconds to
# wait before retrying a request.
RETRY_WAIT_EXPONENTIAL_MAX = 10000  # Maximum number of milliseconds to wait
# before retrying a request.
CW_DEFAULT_PAGE = 1  # CW Pagination is 1-indexed
CONTENT_DISPOSITION_RE = re.compile(
    '^attachment; filename=\"{0,1}(.*?)\"{0,1}$'
)

logger = logging.getLogger(__name__)


class ConnectWiseAPIError(Exception):
    """Raise this, not request exceptions."""
    pass


class ConnectWiseAPIClientError(ConnectWiseAPIError):
    """
    Raise this to indicate any http error that falls within the
    4xx class of http status codes.
    """
    pass


class ConnectWiseAPIServerError(ConnectWiseAPIError):
    """
    Raise this to indicate a Server Error
    https://developer.connectwise.com/Manage/Developer_Guide#HTTP_Response_Codes
    500 class of http status codes.
    """
    pass


class ConnectWiseRecordNotFoundError(ConnectWiseAPIClientError):
    """The record was not found."""
    pass


class ConnectWiseSecurityPermissionsException(ConnectWiseAPIClientError):
    """The API credentials have insufficient security permissions."""
    pass


class CompanyInfoManager:
    COMPANYINFO_ENDPOINT = '{}/login/companyinfo/{}'

    def get_company_info(self, server_url, company_id):
        company_endpoint = self.COMPANYINFO_ENDPOINT.format(
            server_url, company_id
        )

        try:
            logger.debug('Making GET request to {}'.format(company_endpoint))
            response = requests.get(company_endpoint)
            if 200 <= response.status_code < 300:
                resp_json = response.json()
                if resp_json is None:
                    # CW returns None if company ID is unknown.
                    raise ConnectWiseAPIError(
                        'Null response received- company ID may be unknown.'
                    )
                else:
                    return resp_json
            elif response.status_code == 404:
                msg = 'CompanyInfo not found: {}'.format(company_endpoint)
                raise ConnectWiseRecordNotFoundError(msg)
            else:
                raise ConnectWiseAPIError(response.content)
        except requests.RequestException as e:
            raise ConnectWiseAPIError(
                'CompanyInfo request failed: GET {}: {}'.format(
                    company_endpoint, e
                )
            )

    def fetch_api_codebase(self, server_url, company_id, force_fetch=True):
        """
        Returns the Codebase value for the hosted Connectwise instance
        at the supplied URL. The Codebase is retrieved from the cache
        or, if it is not found, it is retrieved from the companyinfo endpoint.
        """
        cache_key = 'api_codebase'
        codebase_result = DEFAULT_CW_API_CODEBASE
        codebase_updated = False

        if CW_CLOUD_DOMAIN in server_url:
            logger.debug('Fetching ConnectWise codebase value from cache.')
            codebase_from_cache = cache.get(cache_key)
            logger.info(
                'Cached ConnectWise codebase was: {}'.format(
                    codebase_from_cache
                )
            )
            if not codebase_from_cache or force_fetch:
                company_info_json = self.get_company_info(
                    server_url, company_id
                )
                codebase_from_api = company_info_json['Codebase']

                codebase_updated = codebase_from_cache != codebase_from_api

                codebase_result = codebase_from_api
                logger.info(
                    'Setting ConnectWise codebase cache value to: {}'.format(
                        codebase_result
                    )
                )
                cache.set(cache_key, codebase_result)
            else:
                codebase_result = codebase_from_cache
        return codebase_result, codebase_updated


def retry_if_api_error(exception):
    """
    Return True if we should retry (in this case when it's an
    ConnectWiseAPIError), False otherwise.

    Basically, don't retry on ConnectWiseAPIClientError, because those are the
    type of exceptions where retrying won't help (404s, 403s, etc).
    """
    return type(exception) is ConnectWiseAPIError


class ConnectWiseAPIClient(object):
    API = None
    MAX_404_ATTEMPTS = 1

    def __init__(
        self,
        company_id=None,
        server_url=None,
        client_id=None,
        api_public_key=None,
        api_private_key=None,
    ):
        if not company_id:
            company_id = settings.CONNECTWISE_CREDENTIALS['company_id']
        if not server_url:
            server_url = settings.CONNECTWISE_SERVER_URL
        if not client_id:
            client_id = settings.CONNECTWISE_CLIENTID
        if not api_public_key:
            api_public_key = settings.CONNECTWISE_CREDENTIALS['api_public_key']
        if not api_private_key:
            api_private_key = settings.CONNECTWISE_CREDENTIALS[
                'api_private_key'
            ]

        if not self.API:
            raise ValueError('API not specified')

        self.info_manager = CompanyInfoManager()
        self.company_id = company_id
        self.api_public_key = api_public_key
        self.api_private_key = api_private_key
        self.server_url = self.change_cw_cloud_url(server_url)
        self.client_id = client_id
        self.auth = (
            '{0}+{1}'.format(company_id, self.api_public_key),
            '{0}'.format(self.api_private_key),
        )

        self.request_settings = DjconnectwiseSettings().get_settings()
        self.timeout = self.request_settings['timeout']

    def _endpoint(self, path):
        return '{0}{1}'.format(self.api_base_url, path)

    def _log_failed(self, response):
        logger.error('Failed API call: {0} - {1} - {2}'.format(
            response.url, response.status_code, response.content))

    def _prepare_error_response(self, response):
        error = response.content.decode("utf-8")
        # decode the bytes encoded error to a string
        # error = error.args[0].decode("utf-8")
        error = error.replace('\r\n', '')
        messages = []

        try:
            error = json.loads(error)
            stripped_message = error.get('message').rstrip('.') if \
                error.get('message') else 'No message'
            primary_error_msg = '{}.'.format(stripped_message)
            if error.get('errors'):
                for error_message in error.get('errors'):
                    messages.append(
                        '{}.'.format(error_message.get('message').rstrip('.'))
                    )

            messages = ' The error was: '.join(messages)

            msg = '{} {}'.format(primary_error_msg, messages)

        except json.decoder.JSONDecodeError:
            # JSON decoding failed
            msg = 'An error occurred: {} {}'.format(response.status_code,
                                                    error)
        except KeyError:
            # 'code' or 'message' was not found in the error
            msg = 'An error occurred: {} {}'.format(response.status_code,
                                                    error)
        return msg

    @property
    def api_base_url(self):
        # This will be for this particular API-
        # i.e. https://connectwise.example.com/v4_6_release/apis/3.0/service/
        api_base_url, _ = self.build_api_base_url()
        return api_base_url

    def build_api_base_url(self, force_fetch=False):
        api_codebase, codebase_updated = \
            self.info_manager.fetch_api_codebase(
                self.server_url, self.company_id, force_fetch=force_fetch
            )

        api_base_url = '{0}/{1}apis/3.0/{2}/'.format(
            self.server_url,
            api_codebase,
            self.API,
        )

        return api_base_url, codebase_updated

    def get_headers(self):
        headers = {}

        response_version = self.request_settings.get('response_version')

        if response_version:
            headers['Accept'] = \
                'application/vnd.connectwise.com+json; version={}' \
                .format(response_version)

        if self.client_id:
            headers['clientId'] = self.client_id

        return headers

    def fetch_resource(self, endpoint_url, params=None, should_page=False,
                       retry_counter=None,
                       *args, **kwargs):
        """
        Issue a GET request to the specified REST endpoint.

        retry_counter is a dict in the form {'count': 0} that is passed in
        to verify the number of attempts that were made.
        """
        @retry(stop_max_attempt_number=self.request_settings['max_attempts'],
               wait_exponential_multiplier=RETRY_WAIT_EXPONENTIAL_MULTAPPLIER,
               wait_exponential_max=RETRY_WAIT_EXPONENTIAL_MAX,
               retry_on_exception=retry_if_api_error)
        def _fetch_resource(endpoint_url, params=None, should_page=False,
                            retry_counter=None, *args, **kwargs):
            if not retry_counter:
                retry_counter = {'count': 0}
            retry_counter['count'] += 1

            if not params:
                params = {}

            try:
                endpoint = self._endpoint(endpoint_url)
                logger.debug('Making GET request to {}'.format(endpoint))

                conditions_str = ''
                conditions = kwargs.get('conditions')
                if conditions:
                    logger.debug('Conditions: {}'.format(conditions))
                    conditions_str = 'conditions={}'.format(
                        self.prepare_conditions(conditions)
                    )
                    # URL encode needed characters
                    conditions_str = conditions_str.replace("+", "%2B")
                    conditions_str = conditions_str.replace(" ", "+")

                if should_page:
                    params['pageSize'] = kwargs.get('page_size',
                                                    CW_RESPONSE_MAX_RECORDS)
                    params['page'] = kwargs.get('page', CW_DEFAULT_PAGE)

                    delimiter = '' if endpoint_url[-1] == '&' else '?'

                    endpoint += "{}pageSize={}&page={}".format(
                        delimiter, params['pageSize'], params['page'])

                    endpoint += "&" + conditions_str
                else:
                    endpoint += "?" + conditions_str

                response = requests.get(
                    endpoint,
                    auth=self.auth,
                    timeout=self.timeout,
                    headers=self.get_headers(),
                )
                logger.debug(" URL: {}".format(response.url))

            except requests.RequestException as e:
                logger.error('Request failed: GET {}: {}'.format(endpoint, e))
                raise ConnectWiseAPIError('{}'.format(e))

            if 200 <= response.status_code < 300:
                try:
                    return response.json()
                except JSONDecodeError as e:
                    logger.error(
                        'Request failed during decoding JSON: GET {}: {}'
                        .format(endpoint, e)
                    )
                    raise ConnectWiseAPIError('JSONDecodeError: {}'.format(e))

            elif response.status_code == 404:
                msg = 'Resource not found: {}'.format(response.url)
                logger.warning(msg)
                # If this is the first failure, try updating the
                # company info codebase value and let it be retried by the
                # @retry decorator.
                if retry_counter['count'] <= self.MAX_404_ATTEMPTS:
                    _, codebase_updated = self.build_api_base_url(
                        force_fetch=True
                    )
                    if codebase_updated:
                        # Since the codebase was updated, it is worthwhile
                        # to try this request again. It could be that the 404
                        # was due to hosted ConnectWise changing the codebase
                        # URL recently. So raise ConnectWiseAPIError, which
                        # will cause the call to be retried.
                        logger.info('Codebase value has changed, so this '
                                    'request will be retried.')
                        raise ConnectWiseAPIError(response.content)
                raise ConnectWiseRecordNotFoundError(msg)

            elif response.status_code == 403:
                self._log_failed(response)
                raise ConnectWiseSecurityPermissionsException(
                    self._prepare_error_response(response),
                    response.status_code
                )

            elif 400 <= response.status_code < 499:
                self._log_failed(response)
                raise ConnectWiseAPIClientError(
                    self._prepare_error_response(response))
            elif response.status_code == 500:
                self._log_failed(response)
                raise ConnectWiseAPIServerError(
                    self._prepare_error_response(response))
            else:
                self._log_failed(response)
                raise ConnectWiseAPIError(
                    self._prepare_error_response(response))

        if not retry_counter:
            retry_counter = {'count': 0}
        return _fetch_resource(endpoint_url, params=params,
                               should_page=should_page,
                               retry_counter=retry_counter,
                               *args, **kwargs)

    def prepare_conditions(self, conditions):
        """
        From the given array of individual conditions, format the conditions
        URL parameter according to ConnectWise requirements.
        """
        return '({})'.format(' and '.join(conditions))

    def request(self, method, endpoint_url, body=None):
        """
        Issue the given type of request to the specified REST endpoint.
        """
        try:
            logger.debug(
                'Making {} request to {}'.format(method, endpoint_url)
            )
            response = requests.request(
                method,
                endpoint_url,
                json=body,
                auth=self.auth,
                timeout=self.timeout,
                headers=self.get_headers(),
            )
        except requests.RequestException as e:
            logger.error(
                'Request failed: {} {}: {}'.format(method, endpoint_url, e)
            )
            raise ConnectWiseAPIError('{}'.format(e))

        if response.status_code == 204:  # No content
            return None
        elif 200 <= response.status_code < 300:
            return response.json()
        elif response.status_code == 403:
            self._log_failed(response)
            raise ConnectWiseSecurityPermissionsException(
                self._prepare_error_response(response), response.status_code)
        elif response.status_code == 404:
            msg = 'Resource not found: {}'.format(response.url)
            logger.warning(msg)
            raise ConnectWiseRecordNotFoundError(msg)
        elif 400 <= response.status_code < 499:
            self._log_failed(response)
            raise ConnectWiseAPIClientError(
                self._prepare_error_response(response))
        elif response.status_code == 500:
            self._log_failed(response)
            raise ConnectWiseAPIServerError(
                self._prepare_error_response(response))
        else:
            self._log_failed(response)
            raise ConnectWiseAPIError(response)

    def change_cw_cloud_url(self, server_url):
        """
        Replace the user-facing CW CLoud URLs with the API URLs.

        i.e. https://na.myconnectwise.net becomes
        https://api-na.myconnectwise.net

        See https://developer.connectwise.com/Products/Manage/Developer_Guide#Authentication  # noqa
        """
        url = urlparse(server_url)
        if url.netloc not in CW_CLOUD_URLS:
            # Don't change anything, just return.
            return server_url

        return url._replace(netloc=CW_CLOUD_URLS[url.netloc]).geturl()

    def update_instance(self, changed_fields, endpoint_url):
        # Yeah, this schema is a bit bizarre. See CW docs at
        # https://developer.connectwise.com/Manage/Developer_Guide#Patch
        body = self._format_patch_request_body(changed_fields)
        return self.request('patch', endpoint_url, body)

    def create(self, endpoint_url, fields):
        body = self._format_post_request_body(fields)
        return self.request('post', endpoint_url, body)

    def delete(self, endpoint_url):
        return self.request('delete', endpoint_url)

    def _format_patch_request_body(self, changed_fields):
        body = []

        for field, value in changed_fields.items():

            field_update = {
                'op': 'replace',
                'path': field,
            }

            if isinstance(value, datetime.datetime):
                field_update.update({
                    'value': value.astimezone(
                        pytz.timezone('UTC')).strftime(
                            "%Y-%m-%dT%H:%M:%SZ")
                })

            elif isinstance(value, models.Model):
                field_update.update({
                    'value': {
                        'id': value.id,
                    }
                })
            else:
                field_update.update({'value': str(value) if value else ''})

            body.append(field_update)

        return body

    def _format_patch_body(self, **kwargs):
        """
        Formats patch requests for dummy synchronizers to match CWs bizarre
        schema:
        # https://developer.connectwise.com/Manage/Developer_Guide#Patch
        """
        body = []
        for path, value in kwargs.items():
            body.append({
                'op': 'replace',
                'path': path,
                'value': value
            })

        return body

    def _format_post_request_body(self, fields):
        body = {}

        for field, value in fields.items():

            if isinstance(value, datetime.datetime):
                value = value.astimezone(
                    pytz.timezone('UTC')).strftime(
                    "%Y-%m-%dT%H:%M:%SZ")
            elif isinstance(value, models.Model):
                value = {'id': value.id}
            else:
                value = str(value) if value else ''

            body[field] = value

        return body


class CompanyAPIClient(ConnectWiseAPIClient):
    API = 'company'
    ENDPOINT_COMPANIES = 'companies'
    ENDPOINT_CONTACTS = 'contacts'
    ENDPOINT_COMPANY_STATUSES = '{}/statuses'.format(ENDPOINT_COMPANIES)
    ENDPOINT_COMPANY_TYPES = '{}/types'.format(ENDPOINT_COMPANIES)
    ENDPOINT_COMPANY_COMMUNICATION_TYPES = 'communicationTypes'
    ENDPOINT_CONTACT_COMMUNICATIONs = 'communications'
    ENDPOINT_COMPANY_NOTE_TYPES = 'noteTypes'
    ENDPOINT_COMPANY_TEAM = 'teams'
    ENDPOINT_COMPANY_TEAM_ROLE = 'teamRoles'

    def __init__(self, *args, **kwargs):
        # TODO This init method is temporary, remove in 1840 as well as
        #   references to self.communication_type_endpoint
        super().__init__(*args, **kwargs)

        self.communication_type_endpoint = False

    def by_id(self, company_id):
        endpoint_url = '{}/{}'.format(self.ENDPOINT_COMPANIES, company_id)
        return self.fetch_resource(endpoint_url)

    def get_single_contact(self, contact_id):
        endpoint_url = '{}/{}'.format(self.ENDPOINT_CONTACTS, contact_id)
        return self.fetch_resource(endpoint_url)

    def get_single_communication(self, communication_id):
        endpoint_url = '{}/{}/{}'.format(self.ENDPOINT_CONTACTS,
                                         self.ENDPOINT_CONTACT_COMMUNICATIONs,
                                         communication_id)
        return self.fetch_resource(endpoint_url)

    def get_companies(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_COMPANIES, should_page=True,
                                   *args, **kwargs)

    def get_company_statuses(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_COMPANY_STATUSES,
                                   should_page=True,
                                   *args, **kwargs)

    def get_company_types(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_COMPANY_TYPES,
                                   should_page=True,
                                   *args, **kwargs)

    def get_company_note_types(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_COMPANY_NOTE_TYPES,
                                   should_page=True,
                                   *args, **kwargs)

    def get_contacts(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_CONTACTS, should_page=True,
                                   *args, **kwargs)

    def get_contact_communications(self, contact_id, *args, **kwargs):
        endpoint_url = '{}/{}/communications'.format(self.ENDPOINT_CONTACTS,
                                                     contact_id)
        return self.fetch_resource(endpoint_url, should_page=True,
                                   *args, **kwargs)

    def get_communication_types(self, *args, **kwargs):
        self.communication_type_endpoint = True
        return self.fetch_resource(self.ENDPOINT_COMPANY_COMMUNICATION_TYPES,
                                   should_page=True, *args, **kwargs)

    def get_company_team_role(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_COMPANY_TEAM_ROLE,
                                   should_page=True, *args, **kwargs)

    def get_company_team(self, company_id, *args, **kwargs):
        endpoint_url = '{}/{}/{}'.format(self.ENDPOINT_COMPANIES, company_id,
                                         self.ENDPOINT_COMPANY_TEAM)
        return self.fetch_resource(endpoint_url, should_page=True,
                                   *args, **kwargs)

    def get_headers(self):
        """
        This is a temporary fix for setting the response version for the
        communication type endpoint. It does not accept our currently supported
        response version (2019.5). This will be fixed in an upcoming issue
        """
        headers = super().get_headers()

        if self.communication_type_endpoint:
            # Just checking if the response version exists
            response_version = self.request_settings.get('response_version')

            if response_version:
                headers['Accept'] = \
                    'application/vnd.connectwise.com+json; version=2020.4'

        return headers


class ScheduleAPIClient(ConnectWiseAPIClient):
    API = 'schedule'
    ENDPOINT_ENTRIES = 'entries'
    ENDPOINT_SCHEDULE_TYPES = 'types'
    ENDPOINT_SCHEDULE_STATUSES = 'statuses'
    ENDPOINT_CALENDARS = 'calendars'
    ENDPOINT_HOLIDAY = 'holidayLists'

    def get_schedule_statuses(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_SCHEDULE_STATUSES,
                                   should_page=True,
                                   *args, **kwargs)

    def get_schedule_types(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_SCHEDULE_TYPES,
                                   should_page=True,
                                   *args, **kwargs)

    def get_schedule_entries(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_ENTRIES,
                                   should_page=True,
                                   *args, **kwargs)

    def get_schedule_entry(self, entry_id):
        endpoint_url = '{}/{}'.format(self.ENDPOINT_ENTRIES, entry_id)
        return self.fetch_resource(endpoint_url)

    def post_schedule_entry(self, target, schedule_type, **kwargs):
        endpoint_url = self._endpoint(self.ENDPOINT_ENTRIES)

        body = {
                    "objectId": target.id,
                    "type": {
                        "id": schedule_type.id,
                        "identifier": schedule_type.identifier,
                    }
                }

        member = kwargs.get("member")
        if member:
            body.update({
                "member": {
                            "id": member.id,
                            "identifier": member.identifier,
                            "name": str(member)
                          }
            })

        date_start = kwargs.get("date_start")
        if date_start:
            body["dateStart"] = date_start.astimezone(
                pytz.timezone('UTC')).strftime("%Y-%m-%dT%H:%M:%SZ")

        date_end = kwargs.get("date_end")
        if date_end:
            body["dateEnd"] = kwargs.get("date_end")
            body["dateEnd"] = date_end.astimezone(
                pytz.timezone('UTC')).strftime("%Y-%m-%dT%H:%M:%SZ")

        allow_conflicts = kwargs.get("allow_conflicts")
        if allow_conflicts is not None:
            body["allowScheduleConflictsFlag"] = allow_conflicts

        return self.request('post', endpoint_url, body)

    def patch_schedule_entry(self, **kwargs):
        schedule_id = kwargs.get("id")
        endpoint_url = self._endpoint(
            "{}/{}".format(self.ENDPOINT_ENTRIES, schedule_id))

        # Yeah, this schema is a bit bizarre. See CW docs at
        # https://developer.connectwise.com/Manage/Developer_Guide#Patch
        body = [
            {
                'op': 'replace',
                'path': 'doneFlag',
                'value': kwargs.get("done_flag")
            }
        ]
        return self.request('patch', endpoint_url, body)

    def delete_schedule_entry(self, entry_id):
        endpoint_url = self._endpoint(
            '{}/{}'.format(self.ENDPOINT_ENTRIES, entry_id))
        return requests.request(
            'delete',
            endpoint_url,
            auth=self.auth,
            timeout=self.timeout,
            headers=self.get_headers(),
        )

    def get_calendars(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_CALENDARS,
                                   should_page=True,
                                   *args, **kwargs)

    def get_holidays(self, holiday_list_id, *args, **kwargs):
        endpoint_url = '{}/{}/holidays'.format(
            self.ENDPOINT_HOLIDAY,
            holiday_list_id
            )
        return self.fetch_resource(
            endpoint_url,
            should_page=True,
            *args,
            **kwargs
            )

    def get_holiday_lists(self, *args, **kwargs):
        return self.fetch_resource(
            self.ENDPOINT_HOLIDAY,
            should_page=True,
            *args,
            **kwargs
            )


class TimeAPIClient(ConnectWiseAPIClient):
    API = 'time'
    ENDPOINT_ENTRIES = 'entries'
    ENDPOINT_WORK_TYPES = 'workTypes'
    ENDPOINT_WORK_ROLES = 'workRoles'

    def get_time_entries(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_ENTRIES,
                                   should_page=True,
                                   *args, **kwargs)

    def get_work_types(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_WORK_TYPES,
                                   should_page=True,
                                   *args, **kwargs)

    def get_work_roles(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_WORK_ROLES,
                                   should_page=True,
                                   *args, **kwargs)

    def post_time_entry(self, target_data, **kwargs):
        endpoint_url = self._endpoint(self.ENDPOINT_ENTRIES)

        time_start = kwargs.get("time_start")
        time_start = time_start.astimezone(pytz.timezone('UTC')).strftime(
            "%Y-%m-%dT%H:%M:%SZ")

        body = {
                    "chargeToId": target_data['id'],
                    "chargeToType": target_data['type'],
                    "timeStart": time_start,
                    "addToDetailDescriptionFlag": kwargs
                    .get("description_flag"),
                    "addToInternalAnalysisFlag": kwargs
                    .get("analysis_flag"),
                    "addToResolutionFlag": kwargs.get("resolution_flag"),
                    "emailResourceFlag": kwargs.get("resource_flag"),
                    "emailContactFlag": kwargs.get("contact_flag"),
                    "emailCcFlag": kwargs.get("cc_flag"),
                }

        member = kwargs.get("resource")
        if member:
            body.update({
                "member": {
                    "id": member.id,
                    "identifier": member.identifier,
                    "name": str(member),
                }

            })

        time_end = kwargs.get("time_end")
        if time_end:
            time_end = time_end.astimezone(pytz.timezone('UTC')).strftime(
                "%Y-%m-%dT%H:%M:%SZ")
            body.update({"timeEnd": time_end})

        hours_deduct = kwargs.get("hours_deduct")
        if hours_deduct:
            body.update({"hoursDeduct": hours_deduct})

        actual_hours = kwargs.get("actual_hours")
        if actual_hours:
            body.update({"actualHours": actual_hours})

        billable_option = kwargs.get("billable_option")
        if billable_option:
            body.update({"billableOption": billable_option})

        work_type = kwargs.get("work_type")
        if work_type:
            body.update({
                "workType": {
                    "name": str(work_type)
                }
            })

        work_role = kwargs.get("work_role")
        if work_role:
            body.update({
                "workRole": {
                    "name": str(work_role)
                }
            })

        agreement = kwargs.get("agreement")
        if agreement:
            body.update({
                "agreement": {
                    "id": agreement.id,
                    "name": str(agreement)
                }
            })

        notes = kwargs.get("notes")
        if notes:
            body.update({"notes": notes})

        email_cc = kwargs.get("email_cc")
        if email_cc:
            body.update({
                "emailCc": email_cc
            })

        return self.request('post', endpoint_url, body)

    def update_time_entry(self, time_entry):
        endpoint_url = self._endpoint(
            '{}/{}'.format(self.ENDPOINT_ENTRIES, time_entry.id)
        )
        body = [{
            'op': 'replace',
            'path': 'notes',
            'value': time_entry.notes
        }]
        return self.request('patch', endpoint_url, body)


class SalesAPIClient(ConnectWiseAPIClient):
    API = 'sales'
    ENDPOINT_OPPORTUNITIES = 'opportunities'
    ENDPOINT_OPPORTUNITY_STATUSES = \
        '{}/statuses'.format(ENDPOINT_OPPORTUNITIES)
    ENDPOINT_OPPORTUNITY_TYPES = \
        '{}/types'.format(ENDPOINT_OPPORTUNITIES)
    ENDPOINT_ACTIVITIES = 'activities'
    ENDPOINT_ACTIVITY_STATUSES = '{}/statuses'.format(ENDPOINT_ACTIVITIES)
    ENDPOINT_ACTIVITY_TYPES = '{}/types'.format(ENDPOINT_ACTIVITIES)
    ENDPOINT_PROBABILITIES = 'probabilities'
    ENDPOINT_STAGES = 'stages'

    def by_id(self, opportunity_id):
        endpoint_url = '{}/{}'.format(
            self.ENDPOINT_OPPORTUNITIES, opportunity_id)
        return self.fetch_resource(endpoint_url)

    def get_activities(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_ACTIVITIES,
                                   should_page=True,
                                   *args, **kwargs)

    def get_single_activity(self, activity_id):
        endpoint_url = '{}/{}'.format(
            self.ENDPOINT_ACTIVITIES, activity_id)
        return self.fetch_resource(endpoint_url)

    def get_activity_statuses(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_ACTIVITY_STATUSES,
                                   should_page=True,
                                   *args, **kwargs)

    def get_activity_types(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_ACTIVITY_TYPES,
                                   should_page=True,
                                   *args, **kwargs)

    def get_probabilities(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_PROBABILITIES,
                                   should_page=True,
                                   *args, **kwargs)

    def get_opportunities(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_OPPORTUNITIES,
                                   should_page=True,
                                   *args, **kwargs)

    def get_opportunity_statuses(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_OPPORTUNITY_STATUSES,
                                   should_page=True,
                                   *args, **kwargs)

    def get_opportunity_types(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_OPPORTUNITY_TYPES,
                                   should_page=True,
                                   *args, **kwargs)

    def get_opportunity_stages(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_STAGES,
                                   should_page=True,
                                   *args, **kwargs)

    def get_notes(self, opportunity_id, *args, **kwargs):
        """
        Returns the notes associated with the specific opportunity.
        """
        endpoint_url = '{}/{}/notes'.format(self.ENDPOINT_OPPORTUNITIES,
                                            opportunity_id
                                            )
        return self.fetch_resource(endpoint_url, should_page=True,
                                   *args, **kwargs)

    def update_opportunity(self, obj, changed_fields):
        endpoint_url = self._endpoint(
            '{}/{}'.format(self.ENDPOINT_OPPORTUNITIES, obj.id)
        )
        return self.update_instance(changed_fields, endpoint_url)

    def update_activity(self, obj, changed_fields):
        endpoint_url = self._endpoint(
            '{}/{}'.format(self.ENDPOINT_ACTIVITIES, obj.id)
        )
        return self.update_instance(changed_fields, endpoint_url)


class SystemAPIClient(ConnectWiseAPIClient):
    API = 'system'

    # endpoints
    ENDPOINT_MEMBERS = 'members/'
    ENDPOINT_DOCUMENTS_DOWNLOAD = 'documents/{}/download'
    ENDPOINT_MEMBERS_COUNT = 'members/count'
    ENDPOINT_CALLBACKS = 'callbacks/'
    ENDPOINT_INFO = 'info/'
    # Locations in the system API are actually territories
    ENDPOINT_LOCATIONS = 'locations/'
    ENDPOINT_OTHER = 'myCompany/other/'
    ENDPOINT_DOCUMENTS = 'documents/'

    def get_connectwise_version(self):
        result = self.fetch_resource(self.ENDPOINT_INFO)
        return result.get('version', '')

    def get_members(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_MEMBERS,
                                   should_page=True, *args, **kwargs)

    def get_member_count(self):
        return self.fetch_resource(self.ENDPOINT_MEMBERS_COUNT)

    def get_territories(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_LOCATIONS,
                                   should_page=True, *args, **kwargs)

    def get_callbacks(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_CALLBACKS,
                                   should_page=True, *args, **kwargs)

    def get_mycompanyother(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_OTHER, *args, **kwargs)

    def delete_callback(self, entry_id):
        endpoint = self._endpoint(
            '{}{}'.format(self.ENDPOINT_CALLBACKS, entry_id)
        )
        return self.request(
            'delete',
            endpoint,
        )

    def create_callback(self, callback_entry):
        endpoint = self._endpoint(self.ENDPOINT_CALLBACKS)
        return self.request(
            'post',
            endpoint,
            body=callback_entry,
        )

    def get_member_by_identifier(self, identifier):
        return self.fetch_resource('members/{0}'.format(identifier))

    def get_member_image_by_photo_id(self, photo_id, username):
        """
        Return a (filename, content) tuple.

        This requires this permission:
        Companies => Manage Documents => Inquire Level: All
        """

        filename, response = self.document_download(photo_id)

        if filename and response:
            headers = response.headers
            content_disposition_header = headers.get('Content-Disposition',
                                                     default='')

            logger.info(
                "Got member '{}' image; size {} bytes and "
                "content-disposition header '{}'".format(
                    username,
                    len(response.content),
                    content_disposition_header
                )
            )
        return filename, response.content if response else response

    def _attachment_filename(self, content_disposition):
        """
        Return the attachment filename from the content disposition header.

        If there's no match, return None.
        """
        m = CONTENT_DISPOSITION_RE.match(content_disposition)
        return m.group(1) if m else None

    def document_download(self, document_id):
        endpoint = self._endpoint(
            self.ENDPOINT_DOCUMENTS_DOWNLOAD.format(document_id)
        )

        try:
            logger.debug('Making GET request to {}'.format(endpoint))
            response = requests.get(
                endpoint,
                auth=self.auth,
                timeout=self.timeout,
                headers=self.get_headers(),
            )
        except requests.RequestException as e:
            logger.error('Request failed: GET {}: {}'.format(endpoint, e))
            raise ConnectWiseAPIError('{}'.format(e))

        if 200 <= response.status_code < 300:
            headers = response.headers
            content_disposition_header = headers.get('Content-Disposition',
                                                     default='')

            attachment_filename = self._attachment_filename(
                content_disposition_header)
            return attachment_filename, response
        else:
            self._log_failed(response)
            return None, None

    def get_attachments(self, object_id, *args, **kwargs):
        endpoint_url = \
            f'{self.ENDPOINT_DOCUMENTS}' \
            f'?recordType=Ticket&recordId={object_id}&'
        return self.fetch_resource(endpoint_url,
                                   should_page=True, *args, **kwargs)

    def get_attachment_count(self, object_id):
        endpoint_url = f'{self.ENDPOINT_DOCUMENTS}count' \
                       f'?recordType=Ticket&recordId={object_id}&'
        return self.fetch_resource(endpoint_url).get('count', 0)

    def get_attachment(self, document_id):
        return self.document_download(document_id)


class TicketAPIMixin:
    ENDPOINT_TICKETS = 'tickets'

    def tickets_count(self, **kwargs):
        return self.fetch_resource(
            '{}/count'.format(self.ENDPOINT_TICKETS), **kwargs).get('count', 0)

    def get_ticket(self, ticket_id):
        endpoint_url = '{}/{}'.format(self.ENDPOINT_TICKETS, ticket_id)
        return self.fetch_resource(endpoint_url)

    def delete_ticket(self, ticket_id):
        endpoint_url = '{}/{}'.format(self.ENDPOINT_TICKETS, ticket_id)
        return self.request('delete', self._endpoint(endpoint_url))

    def get_ticket_tasks(self, ticket_id, **kwargs):
        endpoint_url = '{}/{}/tasks'.format(
            self.ENDPOINT_TICKETS, ticket_id)
        return self.fetch_resource(endpoint_url, should_page=True, **kwargs)

    def create_ticket_task(self, ticket_id, **kwargs):
        endpoint_url = '{}/{}/tasks'.format(
            self.ENDPOINT_TICKETS, ticket_id)
        return self.request('post', self._endpoint(endpoint_url), kwargs)

    def update_ticket_task(self, task_id, ticket_id, **kwargs):
        endpoint_url = '{}/{}/tasks/{}'.format(
            self.ENDPOINT_TICKETS, ticket_id, task_id)
        body = self._format_patch_body(**kwargs)
        return self.request('patch', self._endpoint(endpoint_url), body)

    def delete_ticket_task(self, ticket_id, **kwargs):
        endpoint_url = '{}/{}/tasks/{}'.format(
            self.ENDPOINT_TICKETS, ticket_id, kwargs['id'])
        return self.request('delete', self._endpoint(endpoint_url))

    def get_tickets(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_TICKETS, should_page=True,
                                   *args, **kwargs)

    def update_ticket(self, ticket, changed_fields):
        endpoint_url = self._endpoint(
            '{}/{}'.format(self.ENDPOINT_TICKETS, ticket.id)
        )
        body = self._format_ticket_patch_body(changed_fields)
        return self.request('patch', endpoint_url, body)

    def create_ticket(self, fields):
        endpoint_url = self._endpoint(
            '{}/'.format(self.ENDPOINT_TICKETS)
        )
        body = self._format_ticket_post_body(fields)
        return self.request('POST', endpoint_url, body)

    def _format_ticket_post_body(self, fields):
        # CW formats POST and PATCH very differently, thats why there are two
        #  different methods for it.
        # TODO reduce duplication between post/patch formatting after switch
        #  to synchronizers
        body = {}

        for field, value in fields.items():

            if isinstance(value, datetime.datetime):
                value = value.astimezone(
                        pytz.timezone('UTC')).strftime(
                        "%Y-%m-%dT%H:%M:%SZ")
            elif isinstance(value, models.Model):
                value = {'id': value.id}
            else:
                value = str(value) if value else ''

            body[field] = value

        return body

    def _format_ticket_patch_body(self, changed_fields):
        # TODO reduce duplication between post/patch formatting after switch
        #  to synchronizers
        body = []

        for field, value in changed_fields.items():

            field_update = {
                'op': 'replace',
                'path': field,
            }

            if isinstance(value, datetime.datetime):
                field_update.update({
                    'value': value.astimezone(
                        pytz.timezone('UTC')).strftime(
                            "%Y-%m-%dT%H:%M:%SZ")
                })

            elif isinstance(value, models.Model):
                field_update.update({
                    'value': {
                        'id': value.id,
                    }
                })
            else:
                field_update.update({'value': str(value) if value else ''})

            body.append(field_update)

        return body

    def update_note(self, note):
        endpoint_url = self._endpoint(
            '{}/{}/notes/{}'.format(
                self.ENDPOINT_TICKETS, note.ticket.id, note.id)
        )
        body = [{
            'op': 'replace',
            'path': 'text',
            'value': note.text
        }]
        return self.request('patch', endpoint_url, body)


class ServiceAPIClient(TicketAPIMixin, ConnectWiseAPIClient):
    API = 'service'
    ENDPOINT_BOARDS = 'boards'
    ENDPOINT_PRIORITIES = 'priorities'
    ENDPOINT_LOCATIONS = 'locations'
    ENDPOINT_SLAS = 'SLAs'
    ENDPOINT_SOURCES = 'sources'

    def get_notes(self, ticket_id, *args, **kwargs):
        """
        Returns the notes associated with the specific ticket.
        """
        endpoint_url = '{}/{}/notes'.format(self.ENDPOINT_TICKETS, ticket_id)
        return self.fetch_resource(endpoint_url, should_page=True,
                                   *args, **kwargs)

    def post_note(self, target_data, **kwargs):
        endpoint_url = self._endpoint(
            '{}/{}/notes'.format(self.ENDPOINT_TICKETS, target_data['id']))

        body = {
            "detailDescriptionFlag": kwargs.get("description_flag"),
            "internalAnalysisFlag": kwargs.get("analysis_flag"),
            "resolutionFlag": kwargs.get("resolution_flag"),
            "processNotifications": kwargs.get("process_notifications"),
            "customerUpdatedFlag": kwargs.get("customer_updated_flag")
        }

        member = kwargs.get("resource")
        if member:
            body.update({
                "member": {
                    "id": member.id,
                    "identifier": member.identifier,
                    "name": str(member),
                }

            })

        contact = kwargs.get("contact")
        if contact:
            body.update({
                "contact": contact
            })

        text = kwargs.get("text")
        if text:
            body.update({"text": text})

        return self.request('post', endpoint_url, body)

    def post_merge_ticket(self, merge_data, **kwargs):
        parent_id = merge_data.get('parent_ticket_id')
        child_id = merge_data.get('child_ticket_id')
        status = merge_data.get('status')

        endpoint_url = self._endpoint(
            '{}/{}/merge'.format(self.ENDPOINT_TICKETS, parent_id))

        body = {
            "mergeTicketIds": [child_id],
            "status": {
                "id": status.id,
                "name": status.name
            }
        }
        return self.request('post', endpoint_url, body)

    def get_statuses(self, board_id, *args, **kwargs):
        """
        Returns the status types associated with the specified board.
        """
        endpoint_url = '{}/{}/statuses'.format(self.ENDPOINT_BOARDS, board_id)
        return self.fetch_resource(endpoint_url, should_page=True,
                                   *args, **kwargs)

    def get_boards(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_BOARDS, should_page=True,
                                   *args, **kwargs)

    def get_board(self, board_id):
        return self.fetch_resource('{}/{}'.format(
            self.ENDPOINT_BOARDS, board_id)
        )

    def get_priorities(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_PRIORITIES, should_page=True,
                                   *args, **kwargs)

    def get_slapriorities(self, sla_id, *args, **kwargs):
        endpoint_url = '{}/{}/priorities'.format(self.ENDPOINT_SLAS, sla_id)
        return self.fetch_resource(endpoint_url, should_page=True,
                                   *args, **kwargs)

    def get_teams(self, board_id, *args, **kwargs):
        endpoint = '{}/{}/teams/'.format(self.ENDPOINT_BOARDS, board_id)
        return self.fetch_resource(endpoint, should_page=True, *args, **kwargs)

    def get_locations(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_LOCATIONS, should_page=True,
                                   *args, **kwargs)

    def get_slas(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_SLAS, should_page=True,
                                   *args, **kwargs)

    def get_types(self, board_id, *args, **kwargs):
        endpoint_url = '{}/{}/types/'.format(self.ENDPOINT_BOARDS, board_id)
        return self.fetch_resource(endpoint_url, should_page=True,
                                   *args, **kwargs)

    def get_subtypes(self, board_id, *args, **kwargs):
        endpoint_url = '{}/{}/subtypes/'.format(self.ENDPOINT_BOARDS, board_id)
        return self.fetch_resource(endpoint_url, should_page=True,
                                   *args, **kwargs)

    def get_items(self, board_id, *args, **kwargs):
        endpoint_url = '{}/{}/items/'.format(self.ENDPOINT_BOARDS, board_id)
        return self.fetch_resource(endpoint_url, should_page=True,
                                   *args, **kwargs)

    def get_type_subtype_item_associations(self, board_id, *args, **kwargs):
        endpoint_url = '{}/{}/typeSubTypeItemAssociations/'.format(
            self.ENDPOINT_BOARDS, board_id)
        return self.fetch_resource(endpoint_url, should_page=True,
                                   *args, **kwargs)

    def get_sources(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_SOURCES, should_page=True,
                                   *args, **kwargs)


class ProjectAPIClient(TicketAPIMixin, ConnectWiseAPIClient):
    API = 'project'
    ENDPOINT_PROJECTS = 'projects/'
    ENDPOINT_PROJECT_STATUSES = 'statuses/'
    ENDPOINT_PROJECT_PHASES = 'phases/'
    ENDPOINT_PROJECT_TYPES = 'projectTypes/'
    ENDPOINT_PROJECT_TEAM_MEMBERS = 'teamMembers/'
    ENDPOINT_PROJECT_NOTES = 'notes'

    def get_project(self, project_id):
        endpoint_url = '{}/{}'.format(self.ENDPOINT_PROJECTS, project_id)
        return self.fetch_resource(endpoint_url)

    def get_projects(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_PROJECTS, should_page=True,
                                   *args, **kwargs)

    def get_project_statuses(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_PROJECT_STATUSES,
                                   should_page=True,
                                   *args, **kwargs)

    def get_project_types(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_PROJECT_TYPES,
                                   should_page=True,
                                   *args, **kwargs)

    def get_project_phases(self, project_id, *args, **kwargs):
        endpoint_url = '{}{}/{}'.format(
            self.ENDPOINT_PROJECTS, project_id, self.ENDPOINT_PROJECT_PHASES
        )
        return self.fetch_resource(endpoint_url,
                                   should_page=True,
                                   *args, **kwargs)

    def get_project_team_members(self, project_id, *args, **kwargs):
        endpoint_url = '{}{}/{}'.format(self.ENDPOINT_PROJECTS, project_id,
                                        self.ENDPOINT_PROJECT_TEAM_MEMBERS)
        return self.fetch_resource(endpoint_url,
                                   should_page=True,
                                   *args, **kwargs)

    def create_project(self, fields):
        endpoint_url = self._endpoint(self.ENDPOINT_PROJECTS)

        return self.create(endpoint_url, fields)

    def update_project(self, project, changed_fields):
        endpoint_url = self._endpoint(
            '{}{}'.format(self.ENDPOINT_PROJECTS, project.id)
        )
        return self.update_instance(changed_fields, endpoint_url)

    def get_project_notes(self, project_id, *args, **kwargs):
        endpoint_url = '{}{}/{}'.format(self.ENDPOINT_PROJECTS, project_id,
                                        self.ENDPOINT_PROJECT_NOTES)
        return self.fetch_resource(endpoint_url, should_page=True,
                                   *args, **kwargs)

    def get_project_notes_count(self, project_id):
        endpoint_url = '{}{}/{}'.format(self.ENDPOINT_PROJECTS, project_id,
                                        self.ENDPOINT_PROJECT_NOTES)
        res = self.fetch_resource(endpoint_url)

        return len(res)


class FinanceAPIClient(ConnectWiseAPIClient):
    API = 'finance'
    ENDPOINT_AGREEMENTS = 'agreements'

    def get_agreements(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_AGREEMENTS, should_page=True,
                                   *args, **kwargs)


class HostedAPIClient(SystemAPIClient):
    ENDPOINT_HOSTED_SETUPS = 'connectwisehostedsetups/'

    def post_hosted_tab(self, *args, **kwargs):
        endpoint_url = self._endpoint(self.ENDPOINT_HOSTED_SETUPS)

        body = {
            'screenId': kwargs.get('screen_id'),
            'description': kwargs.get('description'),
            'url': kwargs.get('url'),
            'origin': kwargs.get('origin'),
            'type': kwargs.get('type')
        }

        return self.request('post', endpoint_url, body)

    def get_hosted_tabs(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_HOSTED_SETUPS,
                                   should_page=True,
                                   *args, **kwargs)

    def delete_hosted_tab(self, hosted_tab_id):
        endpoint_url = self._endpoint(
            '{}/{}'.format(self.ENDPOINT_HOSTED_SETUPS, hosted_tab_id)
        )

        return requests.request('delete', endpoint_url, auth=self.auth)


class HostedReportAPIClient(ConnectWiseAPIClient):
    API = 'system/reports'
    ENDPOINT_HOSTED_REPORT = 'ConnectWiseHostedApiScreen/'

    def get_hosted_screen_ids(self, *args, **kwargs):
        return self.fetch_resource(self.ENDPOINT_HOSTED_REPORT,
                                   *args, **kwargs)



# Generated by Django 4.0.7 on 2023-03-08 17:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0168_ticket_contact_email_address_ticket_contact_name_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='connectwiseboard',
            name='time_entry_discussion_flag',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='connectwiseboard',
            name='time_entry_internal_analysis_flag',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='connectwiseboard',
            name='time_entry_resolution_flag',
            field=models.BooleanField(default=False),
        ),
    ]



# Generated by Django 4.0.10 on 2023-05-19 14:20

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0172_companynotetypetracker'),
    ]

    operations = [
        migrations.RenameField(
            model_name='companynotetype',
            old_name='defaultFlag',
            new_name='default_flag',
        ),
    ]



# Generated by Django 4.0.7 on 2022-10-07 10:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0167_alter_ticket_is_issue_flag'),
    ]

    operations = [
        migrations.AddField(
            model_name='ticket',
            name='contact_email_address',
            field=models.CharField(blank=True, max_length=250, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='contact_name',
            field=models.CharField(blank=True, max_length=62, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='contact_phone_extension',
            field=models.CharField(blank=True, max_length=15, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='contact_phone_number',
            field=models.CharField(blank=True, max_length=20, null=True),
        ),
    ]



# Generated by Django 3.1.2 on 2020-12-09 10:33

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0139_auto_20201208_1204'),
    ]

    operations = [
        migrations.AddField(
            model_name='activity',
            name='contact',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='djconnectwise.contact'),
        ),
        migrations.AddField(
            model_name='opportunity',
            name='contact',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='djconnectwise.contact'),
        ),
        migrations.AddField(
            model_name='project',
            name='contact',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='djconnectwise.contact'),
        ),
        migrations.AddField(
            model_name='ticket',
            name='contact',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='djconnectwise.contact'),
        ),
    ]



# Generated by Django 4.0.10 on 2023-06-28 12:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0175_rename_accountmanagerflag_companyteam_account_manager_flag_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='project',
            name='billing_method',
            field=models.CharField(blank=True, choices=[('ActualRates', 'Actual Rates'), ('FixedFee', 'Fixed Fee'), ('NotToExceed', 'Not To Exceed'), ('OverrideRate', 'Override Rate')], max_length=50, null=True),
        ),
    ]



# Generated by Django 2.1.14 on 2020-09-09 11:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0121_syncjob_skipped'),
    ]

    operations = [
        migrations.AddField(
            model_name='projectphase',
            name='wbs_code',
            field=models.CharField(blank=True, max_length=50, null=True),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django_extensions.db.fields


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0023_auto_20170605_0705'),
    ]

    operations = [
        migrations.CreateModel(
            name='OpportunityStage',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, serialize=False, auto_created=True)),
                ('created', django_extensions.db.fields.CreationDateTimeField(verbose_name='created', auto_now_add=True)),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(verbose_name='modified', auto_now=True)),
                ('name', models.CharField(max_length=50)),
            ],
            options={
                'ordering': ('-modified', '-created'),
                'get_latest_by': 'modified',
                'abstract': False,
            },
        ),
    ]



# Generated by Django 2.1 on 2019-07-05 10:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0086_merge_20190507_1402'),
    ]

    operations = [
        migrations.AddField(
            model_name='ticket',
            name='automatic_email_cc_flag',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='ticket',
            name='automatic_email_contact_flag',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='ticket',
            name='automatic_email_resource_flag',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='ticket',
            name='bill_time',
            field=models.CharField(blank=True, choices=[('Billable', 'Billable'), ('DoNotBill', 'Do Not Bill'), ('NoCharge', 'No Charge'), ('NoDefault', 'No Default')], max_length=20, null=True),
        ),
    ]



# Generated by Django 2.0 on 2018-08-14 10:46

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0065_auto_20180809_1124'),
    ]

    operations = [
        migrations.AlterField(
            model_name='sla',
            name='based_on',
            field=models.CharField(choices=[('MyCalendar', 'My Company Calendar'), ('Customer', "Customer's Calendar"), ('AllHours', '24 Hours'), ('Custom', 'Custom Calendar')], default='MyCalendar', max_length=50),
        ),
    ]



# Generated by Django 2.1.14 on 2020-06-24 10:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0118_auto_20200623_1016'),
    ]

    operations = [
        migrations.AddField(
            model_name='project',
            name='scheduled_end',
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='project',
            name='scheduled_start',
            field=models.DateField(blank=True, null=True),
        ),
    ]



# Generated by Django 3.1.2 on 2020-10-15 18:04

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0133_activityudftracker_opportunityudftracker_projectudftracker_ticketudftracker'),
    ]

    operations = [
        migrations.RenameField(
            model_name='activityudf',
            old_name='entryMethod',
            new_name='entry_method',
        ),
        migrations.RenameField(
            model_name='activityudf',
            old_name='numberOfDecimals',
            new_name='number_of_decimals',
        ),
        migrations.RenameField(
            model_name='opportunityudf',
            old_name='entryMethod',
            new_name='entry_method',
        ),
        migrations.RenameField(
            model_name='opportunityudf',
            old_name='numberOfDecimals',
            new_name='number_of_decimals',
        ),
        migrations.RenameField(
            model_name='projectudf',
            old_name='entryMethod',
            new_name='entry_method',
        ),
        migrations.RenameField(
            model_name='projectudf',
            old_name='numberOfDecimals',
            new_name='number_of_decimals',
        ),
        migrations.RenameField(
            model_name='ticketudf',
            old_name='entryMethod',
            new_name='entry_method',
        ),
        migrations.RenameField(
            model_name='ticketudf',
            old_name='numberOfDecimals',
            new_name='number_of_decimals',
        ),
    ]



import datetime
import logging
import re
import urllib

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from django_extensions.db.models import TimeStampedModel
from model_utils import FieldTracker

from . import api

logger = logging.getLogger(__name__)


PRIORITY_RE = re.compile(r'^Priority ([\d]+)')

BILL_TYPES = (
        ('Billable', "Billable"),
        ('DoNotBill', "Do Not Bill"),
        ('NoCharge', "No Charge"),
    )


class UpdateConnectWiseMixin:

    def save(self, *args, **kwargs):
        """
        Save the object.
        If update_cw as a kwarg is True, then update ConnectWise with changes.
        """
        update_cw = kwargs.pop('update_cw', False)
        public_key = kwargs.pop('api_public_key', None)
        private_key = kwargs.pop('api_private_key', None)
        changed_fields = kwargs.pop('changed_fields', None)

        if update_cw and changed_fields:
            self.update_cw(
                api_public_key=public_key,
                api_private_key=private_key,
                changed_fields=changed_fields
            )

        # Ensure save is not run before update_cw returns successfully
        super().save(**kwargs)

    def update_cw(self, **kwargs):
        """
        Send ticket updates to ConnectWise. Accepts a changed_fields
        argument which is a list of fields that have been changed and
        should be updated in CW.

        Update in this context means to update connectwise with new data,
        updates can be PATCH to update existing records, or POST to
        create new records.
        """
        # TODO remove all update_cw calls and related after synchronizers
        #  handle updates to PSA
        api_class = self.api_class

        api_client = api_class(
            api_public_key=kwargs.get('api_public_key'),
            api_private_key=kwargs.get('api_private_key')
        )

        changed_fields = kwargs.get('changed_fields')
        # TODO improve naming when moved to synchronizer
        updated_objects = self.get_changed_values(changed_fields)

        return self._update_cw(api_client, updated_objects)

    def get_changed_values(self, changed_field_keys):
        # TODO remove this method after synchronizers handle updates to PSA
        # Prepare the updated fields to be sent to CW. At this point, any
        # updated fields have been set on the object, but not in local DB yet.
        updated_objects = {}
        if changed_field_keys:
            for field in changed_field_keys:
                updated_objects[field] = getattr(self, field)

        return updated_objects


class InvalidStatusError(Exception):
    pass


class SlaGoalsMixin(object):
    """
    Returns the fields relevant to SLA goals for models with SLA information
    """

    def get_stage_hours(self, stage):
        if stage == 'respond':
            return self.respond_hours
        elif stage == 'plan':
            return self.plan_within
        elif stage == 'resolve':
            return self.resolution_hours
        elif stage == 'waiting':
            return 0
        else:
            return None


class SyncJob(models.Model):
    start_time = models.DateTimeField(null=False)
    end_time = models.DateTimeField(blank=True, null=True)
    entity_name = models.CharField(max_length=100)
    synchronizer_class = models.CharField(max_length=100, blank=True,
                                          null=True)
    added = models.PositiveIntegerField(null=True)
    updated = models.PositiveIntegerField(null=True)
    deleted = models.PositiveIntegerField(null=True)
    skipped = models.PositiveIntegerField(null=True)
    success = models.BooleanField(null=True)
    message = models.TextField(blank=True, null=True)
    sync_type = models.CharField(max_length=32, default='full')

    def duration(self):
        if self.start_time and self.end_time:
            return self.end_time - self.start_time


class AvailableConnectWiseBoardManager(models.Manager):
    """Return only active ConnectWise boards."""
    def get_queryset(self):
        return super().get_queryset().filter(inactive=False)


class ConnectWiseBoard(TimeStampedModel):

    name = models.CharField(max_length=255)
    inactive = models.BooleanField(default=False)
    time_entry_discussion_flag = models.BooleanField(default=False)
    time_entry_resolution_flag = models.BooleanField(default=False)
    time_entry_internal_analysis_flag = models.BooleanField(default=False)
    work_role = models.ForeignKey(
        'WorkRole',
        blank=True,
        null=True,
        on_delete=models.SET_NULL
    )
    work_type = models.ForeignKey(
        'WorkType',
        blank=True,
        null=True,
        on_delete=models.SET_NULL
    )
    bill_time = models.CharField(
        max_length=50,
        choices=BILL_TYPES,
        blank=True, null=True
    )
    project_flag = models.BooleanField(default=False)

    objects = models.Manager()
    available_objects = AvailableConnectWiseBoardManager()

    class Meta:
        ordering = ('name',)
        verbose_name = 'ConnectWise board'

    def __str__(self):
        return self.name

    @property
    def board_statuses(self):
        return BoardStatus.available_objects.filter(board=self)

    def get_closed_status(self):
        """
        Find a closed status on the board. Prefer the status
        called "Closed", if such a one exists.
        """
        try:
            closed_status = self.board_statuses.get(
                name='Closed',
                closed_status=True,
            )
        except BoardStatus.DoesNotExist:
            # Using first this time because there could be many and we
            # still don't know which one it could be, or just get None.
            closed_status = self.board_statuses.filter(
                name__iregex=r'^([[:space:]]|[[:punct:]])*Closed'
                             r'([[:space:]]|[[:punct:]])*$',
                closed_status=True,
            ).first()

            if not closed_status:
                # There's nothing called "Closed", or "Closed" surrounded by
                # special charaters.
                # filter...first returns None if nothing is found again.
                closed_status = self.board_statuses.filter(
                    closed_status=True,
                ).first()
        return closed_status


class AvailableBoardStatusManager(models.Manager):
    """
    Return only statuses whose ConnectWise board is active, and whose
    inactive field is False.
    """
    def get_queryset(self):
        return super().get_queryset().filter(
            board__inactive=False, inactive=False
        )


class BoardStatus(TimeStampedModel):
    """
    Used for looking up the status/board id combination
    """
    CLOSED = 'Closed'

    ESCALATION_STATUSES = (
        ('NotResponded', 'Not Responded'),
        ('Responded', 'Responded'),
        ('ResolutionPlan', 'Resolution Plan'),
        ('Resolved', 'Resolved'),
        ('NoEscalation', 'No Escalation')
    )

    # For comparing Escalation Statuses
    ESCALATION_RANK = dict(
        zip(
            [type[0] for type in ESCALATION_STATUSES],
            [i for i in range(5)]
        )
    )

    name = models.CharField(blank=True, null=True, max_length=250)
    sort_order = models.PositiveSmallIntegerField()
    display_on_board = models.BooleanField()
    inactive = models.BooleanField()
    closed_status = models.BooleanField()
    time_entry_not_allowed = models.BooleanField(null=True)
    board = models.ForeignKey(
        'ConnectWiseBoard', blank=True, null=True, on_delete=models.SET_NULL)
    # Letting escalation_status allow blank/null rather than possibly having
    # and incorrect default value in some edge case
    escalation_status = models.CharField(
        max_length=20, choices=ESCALATION_STATUSES, db_index=True,
        blank=True, null=True
    )

    objects = models.Manager()
    available_objects = AvailableBoardStatusManager()

    class Meta:
        ordering = ('board__name', 'sort_order', 'name')
        verbose_name_plural = 'Board statuses'

    def is_non_escalation_status(self):
        return self.escalation_status == 'NoEscalation'

    def __str__(self):
        return '{}/{}'.format(self.board, self.name)

    def __lt__(self, other):
        if other is None:
            return False
        return self.ESCALATION_RANK.get(self.escalation_status) < \
            self.ESCALATION_RANK.get(other.escalation_status)

    def __gt__(self, other):
        if other is None:
            return False
        return self.ESCALATION_RANK.get(self.escalation_status) > \
            self.ESCALATION_RANK.get(other.escalation_status)

    def get_status_rank(self):
        return self.ESCALATION_RANK.get(self.escalation_status)


class Location(TimeStampedModel):
    name = models.CharField(max_length=30)
    where = models.CharField(max_length=100, blank=True, null=True)

    class Meta:
        ordering = ('name',)

    def __str__(self):
        return self.name


class RegularMemberManager(models.Manager):
    """Return members that aren't API members."""
    def get_queryset(self):
        return super().get_queryset().exclude(license_class='A')


class Member(TimeStampedModel):
    LICENSE_CLASSES = (
        ('F', 'Full license'),
        ('A', 'API license'),
        ('C', 'StreamlineIT license'),
        ('X', 'Subcontractor license'),
    )
    identifier = models.CharField(  # This is the CW username
        max_length=15, blank=False, unique=True
    )
    first_name = models.CharField(max_length=30, blank=False)
    last_name = models.CharField(max_length=30, blank=False, null=True)
    office_email = models.EmailField(null=True, blank=True, max_length=250)
    inactive = models.BooleanField(default=False)
    avatar = models.CharField(
        null=True, blank=True, max_length=250,
        verbose_name=_('Member Avatar'), help_text=_('Member Avatar')
    )
    license_class = models.CharField(
        blank=True, null=True, max_length=20,
        choices=LICENSE_CLASSES, db_index=True
    )
    work_type = models.ForeignKey(
        'WorkType', null=True, on_delete=models.SET_NULL)
    work_role = models.ForeignKey(
        'WorkRole', null=True, on_delete=models.SET_NULL)

    objects = models.Manager()
    regular_objects = RegularMemberManager()

    class Meta:
        ordering = ('first_name', 'last_name')

    def __str__(self):
        return '{} {}'.format(self.first_name,
                              self.last_name if self.last_name else '')

    def get_initials(self):
        name_segs = str(self).split(' ')
        initial = ''
        for seg in name_segs:
            seg = seg.strip()
            initial += seg[:1]

        return initial


class AvailableCompanyManager(models.Manager):
    """Return only companies whose deleted_flag isn't true."""
    def get_queryset(self):
        return super().get_queryset().filter(deleted_flag=False)


class Company(TimeStampedModel):
    name = models.CharField(blank=True, null=True, max_length=250,
                            db_index=True)
    identifier = models.CharField(
        blank=True, null=True, max_length=250)
    phone_number = models.CharField(blank=True, null=True, max_length=250)
    fax_number = models.CharField(blank=True, null=True, max_length=250)
    address_line1 = models.CharField(blank=True, null=True, max_length=250)
    address_line2 = models.CharField(blank=True, null=True, max_length=250)
    city = models.CharField(blank=True, null=True, max_length=250)
    state_identifier = models.CharField(blank=True, null=True, max_length=250)
    zip = models.CharField(blank=True, null=True, max_length=250)
    country = models.CharField(blank=True, null=True, max_length=250)
    website = models.CharField(blank=True, null=True, max_length=250)
    market = models.CharField(blank=True, null=True, max_length=250)
    defaultcontactid = models.IntegerField(blank=True, null=True)
    defaultbillingcontactid = models.IntegerField(blank=True, null=True)
    updatedby = models.CharField(blank=True, null=True, max_length=250)
    lastupdated = models.CharField(blank=True, null=True, max_length=250)
    deleted_flag = models.BooleanField(default=False)
    calendar = models.ForeignKey(
        'Calendar',
        blank=True,
        null=True,
        on_delete=models.SET_NULL
        )
    status = models.ForeignKey(
        'CompanyStatus',
        blank=True,
        null=True,
        on_delete=models.SET_NULL
        )
    company_types = models.ManyToManyField('CompanyType')
    territory = models.ForeignKey(
        'Territory',
        blank=True,
        null=True,
        on_delete=models.SET_NULL
    )

    objects = models.Manager()
    available_objects = AvailableCompanyManager()

    class Meta:
        verbose_name_plural = 'companies'
        ordering = ('identifier', )

    def __str__(self):
        return self.name or ''

    def get_identifier(self):
        return self.identifier


class CompanyStatus(models.Model):
    name = models.CharField(max_length=50)
    default_flag = models.BooleanField()
    inactive_flag = models.BooleanField()
    notify_flag = models.BooleanField()
    dissalow_saving_flag = models.BooleanField()
    notification_message = models.CharField(
        max_length=500,
        blank=True,
        null=True
    )
    custom_note_flag = models.BooleanField()
    cancel_open_tracks_flag = models.BooleanField()
    track_id = models.PositiveSmallIntegerField(blank=True, null=True)

    class Meta:
        verbose_name_plural = 'Company statuses'

    def __str__(self):
        return self.name


class CompanyType(models.Model):
    name = models.CharField(max_length=50)
    vendor_flag = models.BooleanField()

    class Meta:
        ordering = ('name', )

    def __str__(self):
        return self.name


class CompanyNoteType(models.Model):
    name = models.CharField(max_length=50)
    identifier = models.CharField(max_length=50)
    default_flag = models.BooleanField()

    class Meta:
        ordering = ('name', )

    def __str__(self):
        return self.name


class CompanyTeam(models.Model):
    company = models.ForeignKey('Company', blank=True,
                                null=True, on_delete=models.CASCADE)
    team_role = models.ForeignKey('CompanyTeamRole',
                                  blank=True, null=True,
                                  on_delete=models.CASCADE)
    location = models.ForeignKey('Location', blank=True,
                                 null=True, on_delete=models.CASCADE)
    contact = models.ForeignKey('Contact', blank=True,
                                null=True, on_delete=models.CASCADE)
    member = models.ForeignKey('Member', blank=True,
                               null=True, on_delete=models.CASCADE)
    account_manager_flag = models.BooleanField()
    tech_flag = models.BooleanField()
    sales_flag = models.BooleanField()

    class Meta:
        ordering = ('id', )

    def __str__(self):
        return self.company.name


class CompanyTeamRole(models.Model):
    name = models.CharField(max_length=50)
    account_manager_flag = models.BooleanField()
    tech_flag = models.BooleanField()
    sales_flag = models.BooleanField()

    class Meta:
        ordering = ('name', )

    def __str__(self):
        return self.name


class Contact(models.Model):

    def __str__(self):
        return '{} {}'.format(self.first_name,
                              self.last_name if self.last_name else '')

    first_name = models.CharField(blank=True, null=True, max_length=200,
                                  db_index=True)
    last_name = models.CharField(blank=True, null=True, max_length=200,
                                 db_index=True)
    title = models.CharField(blank=True, null=True, max_length=200)
    company = models.ForeignKey(
        'Company', null=True, on_delete=models.CASCADE)

    class Meta:
        ordering = ('first_name', 'last_name')

    @property
    def default_email_address(self):
        return self.contactcommunication_set.filter(
            default_flag=True, type_id__email_flag=True).first()

    @property
    def default_phone_number(self):
        return self.contactcommunication_set.filter(
            default_flag=True, type_id__phone_flag=True).first()

    @property
    def default_phone_ext(self):
        phone = self.contactcommunication_set.filter(
            default_flag=True, type_id__phone_flag=True).first()
        return phone.extension if phone else None

    @property
    def default_fax_number(self):
        return self.contactcommunication_set.filter(
            default_flag=True, type_id__fax_flag=True).first()


class ContactCommunication(models.Model):
    contact = models.ForeignKey(
        'Contact', null=True, on_delete=models.CASCADE)
    value = models.CharField(max_length=250)
    extension = models.CharField(blank=True, null=True, max_length=15)
    default_flag = models.BooleanField(blank=True, null=True)
    type = models.ForeignKey(
        'CommunicationType', null=True, on_delete=models.SET_NULL)

    def __str__(self):
        # Do not change this __str__ definition without good reason,
        # email templates rely on it being the value.
        return self.value


class CommunicationType(models.Model):
    description = models.CharField(blank=False, null=False, max_length=250)
    phone_flag = models.BooleanField(default=False)
    fax_flag = models.BooleanField(default=False)
    email_flag = models.BooleanField(default=False)
    default_flag = models.BooleanField(default=False)

    def __str__(self):
        return self.description


class MyCompanyOther(models.Model):
    default_calendar = models.ForeignKey(
        'Calendar',
        blank=True,
        null=True,
        on_delete=models.SET_NULL
        )


class Calendar(models.Model):

    START_TIME = '_start_time'
    END_TIME = '_end_time'

    name = models.CharField(max_length=250)
    holiday_list = models.ForeignKey(
        'HolidayList', on_delete=models.SET_NULL, blank=True, null=True)
    monday_start_time = models.TimeField(auto_now=False, auto_now_add=False,
                                         blank=True, null=True)
    monday_end_time = models.TimeField(auto_now=False, auto_now_add=False,
                                       blank=True, null=True)
    tuesday_start_time = models.TimeField(auto_now=False, auto_now_add=False,
                                          blank=True, null=True)
    tuesday_end_time = models.TimeField(auto_now=False, auto_now_add=False,
                                        blank=True, null=True)
    wednesday_start_time = models.TimeField(auto_now=False, auto_now_add=False,
                                            blank=True, null=True)
    wednesday_end_time = models.TimeField(auto_now=False, auto_now_add=False,
                                          blank=True, null=True)
    thursday_start_time = models.TimeField(auto_now=False, auto_now_add=False,
                                           blank=True, null=True)
    thursday_end_time = models.TimeField(auto_now=False, auto_now_add=False,
                                         blank=True, null=True)
    friday_start_time = models.TimeField(auto_now=False, auto_now_add=False,
                                         blank=True, null=True)
    friday_end_time = models.TimeField(auto_now=False, auto_now_add=False,
                                       blank=True, null=True)
    saturday_start_time = models.TimeField(auto_now=False, auto_now_add=False,
                                           blank=True, null=True)
    saturday_end_time = models.TimeField(auto_now=False, auto_now_add=False,
                                         blank=True, null=True)
    sunday_start_time = models.TimeField(auto_now=False, auto_now_add=False,
                                         blank=True, null=True)
    sunday_end_time = models.TimeField(auto_now=False, auto_now_add=False,
                                       blank=True, null=True)

    def __str__(self):
        return self.name

    def get_day_hours(self, is_start, day):
        if is_start:
            time = self.START_TIME
        else:
            time = self.END_TIME

        days = [
            'monday',
            'tuesday',
            'wednesday',
            'thursday',
            'friday',
            'saturday',
            'sunday'
        ]
        return getattr(self, "{}{}".format(days[day], time), None)

    def get_first_day(self, start):
        """
        For getting the first weekday, and the days until that day, from the
        given start day that has a start and end time.
        """
        start_day = start.weekday()
        days = 0
        while True:
            day = self.get_day_hours(True, start_day)
            if day and \
                    not self.is_holiday(datetime.timedelta(days=days) + start):
                return start_day, days
            start_day = (start_day + 1) % 7
            days += 1
            if days >= 7:
                # Calendar has no hours on any day. This case can also occur
                # if a calendar has seven consecutive holidays.
                return None, None

    def is_holiday(self, date):
        current_day = date.date()
        try:
            holiday = self.holiday_list.holiday_set.filter(date=current_day)
            # Decided to go with filter, and test whether the list is empty or
            # not. Rather than get, and deal with the possibility of
            # DoesNotExist and MultipleObjectsReturned.
        except AttributeError:
            # No holiday list on this calendar
            return False
        if holiday:
            return True
        # Returning False instead of None
        return False

    def get_sla_time(self, start, end):
        # Get the sla-minutes between two dates using the given calendar
        minutes = 0
        day_of_week = start.weekday()
        start_time = datetime.timedelta(hours=start.hour,
                                        minutes=start.minute)

        # Get sla minutes for first day
        start_day_end_time = self.get_day_hours(False, start.weekday())

        if start_day_end_time and \
                not self.is_holiday(timezone.now().astimezone(tz=None)):
            end_of_day = datetime.timedelta(hours=start_day_end_time.hour,
                                            minutes=start_day_end_time.minute)
        else:
            end_of_day = None

        if start.date() == end.date():

            if end_of_day:

                end_time = min(start_day_end_time, end.time())
                end_time_delta = datetime.timedelta(
                    hours=end_time.hour,
                    minutes=end_time.minute
                )

                minutes = (end_time_delta - start_time).total_seconds() / 60
            # return sla time between start and end of day/end time, or zero
            # if start and end was outside of work hours
            return max(minutes, 0)
        else:
            if end_of_day and \
                    not self.is_holiday(timezone.now().astimezone(tz=None)):
                first_day_minutes = (end_of_day - start_time).total_seconds() \
                    / 60
            else:
                first_day_minutes = 0

            sla_minutes = first_day_minutes if first_day_minutes >= 0 else 0
            current = start + datetime.timedelta(days=1)
            day_of_week = (day_of_week + 1) % 7

            while current.date() != end.date():
                start_of_day = self.get_day_hours(True, day_of_week)
                if start_of_day and not self.is_holiday(current):
                    start_of_day = datetime.timedelta(
                        hours=start_of_day.hour,
                        minutes=start_of_day.minute)
                    end_of_day = self.get_day_hours(False, day_of_week)
                    end_of_day = datetime.timedelta(
                        hours=end_of_day.hour,
                        minutes=end_of_day.minute)
                else:
                    # This is a day with no hours, continue to next day
                    day_of_week = (day_of_week + 1) % 7
                    current = current + datetime.timedelta(days=1)
                    continue

                minutes = (end_of_day - start_of_day).total_seconds() / 60
                # 24 hour calendar, minutes are full day
                todays_minutes = minutes if minutes >= 0 else 1440
                sla_minutes = sla_minutes + todays_minutes
                day_of_week = (day_of_week + 1) % 7
                current = current + datetime.timedelta(days=1)

            end_of_day = self.get_day_hours(False, end.weekday())

            if end_of_day:
                end_of_day = datetime.timedelta(
                    hours=end_of_day.hour,
                    minutes=end_of_day.minute)

                end_time = datetime.timedelta(
                    hours=end.hour,
                    minutes=end.minute) if \
                    self.get_day_hours(False, end.weekday()) > \
                    end.time() else end_of_day

                # get sla_minutes for last day
                start_of_day = self.get_day_hours(True, end.weekday())
                start_of_day = datetime.timedelta(hours=start_of_day.hour,
                                                  minutes=start_of_day.minute)

                last_day_minutes = (end_time - start_of_day).total_seconds() \
                    / 60
                minutes = last_day_minutes if last_day_minutes >= 0 else 0
            else:
                minutes = 0
            sla_minutes += minutes
            return sla_minutes

    def next_phase_expiry(self, sla_hours, ticket):

        start = ticket.entered_date_utc.astimezone(tz=None)

        # Start counting from the start of the next business day if the
        # ticket was created on a weekend
        day_of_week, days = self.get_first_day(start)

        if day_of_week is None and days is None:
            ticket.sla_expire_date = None
            return

        sla_minutes, days, minutes, sla_start = self.get_sla_start(
                                               sla_hours,
                                               ticket.minutes_waiting,
                                               start,
                                               day_of_week,
                                               days
                                              )

        # Advance day by day, reducing the sla_minutes by the time in
        # each working day.
        # When minutes goes below zero, add the amount of time left on it
        # that day to the start time of that day, giving you the due date
        while sla_minutes >= 0:
            day_of_week = (day_of_week + 1) % 7
            days += 1

            start_of_day = self.get_day_hours(True, day_of_week)
            if start_of_day and \
                not self.is_holiday(
                                    datetime.timedelta(days=days) + start):
                start_of_day = datetime.timedelta(hours=start_of_day.hour,
                                                  minutes=start_of_day.minute)
                end_of_day = self.get_day_hours(False, day_of_week)
                end_of_day = datetime.timedelta(hours=end_of_day.hour,
                                                minutes=end_of_day.minute)
            else:
                # This is a day with no hours, continue to next day
                continue

            minutes = (end_of_day - start_of_day).total_seconds() / 60
            # 24 hour calendar, minutes are full day
            todays_minutes = minutes if minutes >= 0 else 1440
            sla_minutes = sla_minutes - todays_minutes

        # sla_minutes went below zero so we know that day is the expiry
        # Add the minutes back to sla_minutes and add sla minutes to the
        # start of that day
        sla_minutes = sla_minutes + minutes
        self.set_sla_end(sla_start, days, day_of_week, sla_minutes, ticket)

    def get_sla_start(self, sla_hours, waiting_min, start, day_of_week, days):
        if days > 0:
            start = start + datetime.timedelta(days=days)
            start_of_day = self.get_day_hours(True, day_of_week)
            start = start.replace(
                hour=start_of_day.hour,
                minute=start_of_day.minute
            )
            days = 0

        start_date = datetime.timedelta(hours=start.hour,
                                        minutes=start.minute)

        sla_minutes = (sla_hours * 60) + waiting_min

        end_of_day = self.get_day_hours(False, day_of_week)
        end_of_day = datetime.timedelta(hours=end_of_day.hour,
                                        minutes=end_of_day.minute)
        first_day_minutes = (end_of_day - start_date).total_seconds() / 60

        # If created outside of work hours, take no minutes off and start
        # taking time off next work day
        minutes = first_day_minutes if first_day_minutes >= 0 else 0
        sla_minutes = sla_minutes - minutes

        return sla_minutes, days, minutes, start

    def set_sla_end(self, start, days, day_of_week, sla_minutes, ticket):
        expiry_date = start + datetime.timedelta(days=days)

        if days == 0:
            sla_expire_time = start + datetime.timedelta(minutes=sla_minutes)
            expiry_date = expiry_date.replace(
                hour=sla_expire_time.hour, minute=sla_expire_time.minute)
        else:
            sla_expire_time = self.get_day_hours(True, day_of_week)
            expiry_date = expiry_date.replace(
                hour=sla_expire_time.hour, minute=sla_expire_time.minute) + \
                datetime.timedelta(minutes=sla_minutes)

        ticket.sla_expire_date = expiry_date.astimezone(tz=timezone.utc)


class Holiday(models.Model):
    name = models.CharField(max_length=200)
    all_day_flag = models.BooleanField(default=False)
    date = models.DateField(
                            blank=True,
                            null=True,
                            auto_now=False,
                            auto_now_add=False
                            )
    start_time = models.TimeField(
                                  auto_now=False,
                                  auto_now_add=False,
                                  blank=True,
                                  null=True
                                  )
    end_time = models.TimeField(
                                auto_now=False,
                                auto_now_add=False,
                                blank=True,
                                null=True
                                )
    holiday_list = models.ForeignKey(
        'HolidayList', on_delete=models.CASCADE)

    def __str__(self):
        return self.name


class HolidayList(models.Model):
    name = models.CharField(max_length=200)

    def __str__(self):
        return self.name


class ScheduleType(models.Model):
    name = models.CharField(max_length=50)
    identifier = models.CharField(max_length=1)

    class Meta:
        ordering = ('name', )

    def __str__(self):
        return self.name


class ScheduleStatus(models.Model):
    name = models.CharField(max_length=30)

    class Meta:
        verbose_name_plural = 'Schedule statuses'

    def __str__(self):
        return self.name


class ScheduleEntry(models.Model):
    name = models.CharField(max_length=250, blank=True, null=True)
    date_start = models.DateTimeField(blank=True, null=True)
    date_end = models.DateTimeField(blank=True, null=True)
    done_flag = models.BooleanField(default=False)

    ticket_object = models.ForeignKey(
        'Ticket',
        blank=True,
        null=True,
        on_delete=models.SET_NULL
    )
    activity_object = models.ForeignKey(
        'Activity',
        blank=True,
        null=True,
        on_delete=models.SET_NULL
    )
    member = models.ForeignKey(
        'Member', blank=True, null=True, on_delete=models.SET_NULL)
    where = models.ForeignKey(
        'Location',
        blank=True,
        null=True,
        on_delete=models.SET_NULL
    )
    status = models.ForeignKey(
        'ScheduleStatus',
        blank=True,
        null=True,
        on_delete=models.SET_NULL
    )
    schedule_type = models.ForeignKey(
        'ScheduleType',
        blank=True,
        null=True,
        on_delete=models.SET_NULL
    )

    class Meta:
        verbose_name_plural = 'Schedule entries'
        ordering = ('name', )

    def __str__(self):
        return self.name or ''

    def delete_entry(self, **kwargs):
        """
        Send Delete request to ConnectWise for this entry
        """
        schedule_client = api.ScheduleAPIClient(
            api_public_key=kwargs.get('api_public_key'),
            api_private_key=kwargs.get('api_private_key')
        )
        return schedule_client.delete_schedule_entry(self.id)


class Territory(models.Model):
    name = models.CharField(max_length=250, blank=True, null=True)

    class Meta:
        verbose_name_plural = 'Territories'
        ordering = ('name', )

    def __str__(self):
        return self.name


class UpdateRecordMixin:

    def save(self, *args, **kwargs):
        update_cw = kwargs.pop('update_cw', False)
        public_key = kwargs.pop('api_public_key', None)
        private_key = kwargs.pop('api_private_key', None)

        if update_cw:
            self.update_cw(
                api_public_key=public_key,
                api_private_key=private_key,
            )
        super().save(**kwargs)


class TimeEntry(UpdateRecordMixin, models.Model):
    CHARGE_TYPES = (
        ('ServiceTicket', "Service Ticket"),
        ('ProjectTicket', "Project Ticket"),
        ('ChargeCode', "Charge Code"),
        ('Activity', "Activity")
    )

    class Meta:
        verbose_name_plural = 'Time entries'
        ordering = ('-time_start', 'id')

    def __str__(self):
        return str(self.id) or ''

    actual_hours = models.DecimalField(
        blank=True, null=True, decimal_places=2, max_digits=9)
    billable_option = models.CharField(choices=BILL_TYPES, max_length=250)
    charge_to_type = models.CharField(choices=CHARGE_TYPES, max_length=250)
    hours_deduct = models.DecimalField(
        blank=True, null=True, decimal_places=2, max_digits=9)
    internal_notes = models.TextField(blank=True, null=True, max_length=2000)
    notes = models.TextField(blank=True, null=True, max_length=50000)
    time_start = models.DateTimeField(blank=True, null=True)
    time_end = models.DateTimeField(blank=True, null=True)

    detail_description_flag = models.BooleanField(default=False)
    internal_analysis_flag = models.BooleanField(default=False)
    resolution_flag = models.BooleanField(default=False)

    email_resource_flag = models.BooleanField(default=False)
    email_contact_flag = models.BooleanField(default=False)
    email_cc_flag = models.BooleanField(default=False)
    email_cc = models.CharField(blank=True, null=True, max_length=1000)

    charge_to_id = models.ForeignKey(
        'Ticket', blank=True, null=True, on_delete=models.CASCADE)
    company = models.ForeignKey(
        'Company', blank=True, null=True, on_delete=models.SET_NULL)
    member = models.ForeignKey(
        'Member', blank=True, null=True, on_delete=models.CASCADE)
    work_type = models.ForeignKey(
        'WorkType', blank=True, null=True, on_delete=models.SET_NULL)
    agreement = models.ForeignKey(
        'Agreement', blank=True, null=True, on_delete=models.SET_NULL)

    def get_entered_time(self):
        if self.time_end:
            entered_time = self.time_end
        elif self.actual_hours:
            # timedelta does not like decimals
            minutes = int(self.actual_hours * 60)
            entered_time = self.time_start + timezone.timedelta(
                minutes=minutes)
        else:
            entered_time = self.time_start

        return entered_time

    def get_time_start(self):
        time_start = self.time_start

        # CW allows creating a time entry with actual hours set
        # but no start or end time. In this case, the API returns time_start
        # with the date and time as UTC midnight and time_end as null.
        # Use a naive date instead to prevent conversion problems in the
        # templates.
        if self.time_start and not self.time_end:
            time_start = time_start.date()

        return time_start

    @property
    def note(self):
        return self.notes

    def update_cw(self, **kwargs):

        api_client = api.TimeAPIClient(
            api_public_key=kwargs.get('api_public_key'),
            api_private_key=kwargs.get('api_private_key')
        )
        return api_client.update_time_entry(self)


class AvailableBoardTeamManager(models.Manager):
    """Return only teams whose ConnectWise board is active."""
    def get_queryset(self):
        return super().get_queryset().filter(board__inactive=False)


class Team(TimeStampedModel):
    name = models.CharField(max_length=30)
    board = models.ForeignKey('ConnectWiseBoard', on_delete=models.CASCADE)
    members = models.ManyToManyField('Member')

    objects = models.Manager()
    available_objects = AvailableBoardTeamManager()

    class Meta:
        verbose_name_plural = 'Teams'
        ordering = ('name', 'id')

    def __str__(self):
        return '{}/{}'.format(self.board, self.name)


class TicketPriority(TimeStampedModel):
    name = models.CharField(max_length=50, blank=False)
    # ConnectWise doesn't always return sort and color- not sure why.
    # Sort will be None in this circumstance- dependent code should handle it.
    sort = models.PositiveSmallIntegerField(null=True)
    # Color will be a property that tries to guess at a sensible value.
    _color = models.CharField(
        max_length=50, null=True, blank=True, db_column='color'
    )

    DEFAULT_COLORS = {
        '1': 'red',
        '2': 'orange',
        '3': 'yellow',
        '4': 'white',
        '5': 'darkmagenta',
    }
    DEFAULT_COLOR = 'darkgray'

    class Meta:
        verbose_name_plural = 'ticket priorities'
        ordering = ('sort', 'name', )

    def __str__(self):
        return self.name

    @property
    def color(self):
        """
        If a color has been set, then return it. Otherwise if the name
        matches the common format ("Priority X - ..."), then return
        something sensible based on values seen in the wild.
        """
        if self._color == "Custom":
            return self.DEFAULT_COLOR
        elif self._color:
            return self._color
        else:
            prio_number = None
            prio_match = PRIORITY_RE.match(self.name)
            if prio_match:
                prio_number = prio_match.group(1)
            return TicketPriority.DEFAULT_COLORS.get(
                prio_number,
                TicketPriority.DEFAULT_COLOR
            )

    @color.setter
    def color(self, color):
        self._color = color


class ProjectStatus(TimeStampedModel):
    name = models.CharField(max_length=30)
    default_flag = models.BooleanField(default=False)
    inactive_flag = models.BooleanField(default=False)
    closed_flag = models.BooleanField(default=False)

    class Meta:
        ordering = ('name', )
        verbose_name_plural = 'Project statuses'

    def __str__(self):
        return self.name


class ProjectType(TimeStampedModel):
    name = models.CharField(max_length=30)
    default_flag = models.BooleanField(default=False)
    inactive_flag = models.BooleanField(default=False)

    class Meta:
        ordering = ('name', )
        verbose_name_plural = 'Project types'

    def __str__(self):
        return self.name


class ProjectPhase(TimeStampedModel):
    description = models.CharField(max_length=100)
    scheduled_start = models.DateField(blank=True, null=True)
    scheduled_end = models.DateField(blank=True, null=True)
    actual_start = models.DateField(blank=True, null=True)
    actual_end = models.DateField(blank=True, null=True)

    bill_time = models.CharField(
        max_length=50, choices=BILL_TYPES, blank=True, null=True)
    notes = models.TextField(null=True, blank=False)
    scheduled_hours = models.DecimalField(
        blank=True, null=True, decimal_places=2, max_digits=9)
    actual_hours = models.DecimalField(
        blank=True, null=True, decimal_places=2, max_digits=9)
    budget_hours = models.DecimalField(
        blank=True, null=True, decimal_places=2, max_digits=9)
    wbs_code = models.CharField(max_length=50, blank=True, null=True)

    project = models.ForeignKey(
        'Project', blank=True, null=True, on_delete=models.SET_NULL
    )
    board = models.ForeignKey(
        'ConnectwiseBoard', blank=True, null=True, on_delete=models.SET_NULL
    )

    class Meta:
        verbose_name_plural = 'Project phases'

    def __str__(self):
        return self.description


class ProjectTeamMember(TimeStampedModel):
    start_date = models.DateTimeField(blank=True, null=True)
    end_date = models.DateTimeField(blank=True, null=True)

    project = models.ForeignKey(
        'Project', null=True, on_delete=models.SET_NULL
    )
    member = models.ForeignKey(
        'Member', null=True, on_delete=models.SET_NULL
    )
    work_role = models.ForeignKey(
        'WorkRole', null=True, on_delete=models.SET_NULL
    )

    def __str__(self):
        return '{}/{}'.format(self.id, self.member)


class AvailableProjectManager(models.Manager):
    """
    Return only projects whose status closed field is False.
    """
    def get_queryset(self):
        return super().get_queryset().filter(
            status__closed_flag=False,
        )


class Project(UpdateConnectWiseMixin, TimeStampedModel):
    name = models.CharField(max_length=200)
    actual_hours = models.DecimalField(
        blank=True, null=True, decimal_places=2, max_digits=9)
    budget_hours = models.DecimalField(
        blank=True, null=True, decimal_places=2, max_digits=12)
    scheduled_hours = models.DecimalField(
        blank=True, null=True, decimal_places=2, max_digits=9)
    actual_start = models.DateField(blank=True, null=True)
    actual_end = models.DateField(blank=True, null=True)
    estimated_start = models.DateField(blank=True, null=True)
    estimated_end = models.DateField(blank=True, null=True)
    percent_complete = models.DecimalField(
        blank=True, null=True, decimal_places=2, max_digits=3)
    scheduled_start = models.DateField(blank=True, null=True)
    scheduled_end = models.DateField(blank=True, null=True)
    description = models.TextField(blank=True, null=True)
    udf = models.JSONField(blank=True, null=True)

    ACTUAL_RATES = 'ActualRates'
    FIXED_FEE = 'FixedFee'
    NOT_TO_EXCEED = 'NotToExceed'
    OVERRIDE_RATE = 'OverrideRate'

    BILLING_METHODS = (
        (ACTUAL_RATES, 'Actual Rates'),
        (FIXED_FEE, 'Fixed Fee'),
        (NOT_TO_EXCEED, 'Not To Exceed'),
        (OVERRIDE_RATE, 'Override Rate'),
    )
    billing_method = models.CharField(null=True, blank=True,
                                      choices=BILLING_METHODS, max_length=50)

    board = models.ForeignKey(
        'ConnectWiseBoard',
        blank=True,
        null=True,
        on_delete=models.SET_NULL
    )
    company = models.ForeignKey(
        'Company', on_delete=models.SET_NULL, blank=True, null=True
    )
    contact = models.ForeignKey('Contact', blank=True, null=True,
                                on_delete=models.SET_NULL)
    status = models.ForeignKey(
        'ProjectStatus', blank=True, null=True, on_delete=models.SET_NULL)
    manager = models.ForeignKey(
        'Member',
        blank=True,
        null=True,
        related_name='project_manager',
        on_delete=models.SET_NULL
    )
    type = models.ForeignKey(
        'ProjectType',
        blank=True,
        null=True,
        on_delete=models.SET_NULL
    )
    team_members = models.ManyToManyField(
        'Member', through='ProjectTeamMember',
    )

    objects = models.Manager()
    available_objects = AvailableProjectManager()

    class Meta:
        ordering = ('name', )

    def __str__(self):
        return self.name or ''

    @property
    def api_class(self):
        return api.ProjectAPIClient

    def _update_cw(self, api_client, updated_objects):
        return api_client.update_project(self, updated_objects)

    def get_connectwise_url(self):
        params = dict(
            recordType='ProjectHeaderFV',
            recid=self.id,
            companyName=settings.CONNECTWISE_CREDENTIALS['company_id']
        )
        return '{}/{}?{}'.format(
            settings.CONNECTWISE_SERVER_URL,
            settings.CONNECTWISE_TICKET_PATH,
            urllib.parse.urlencode(params)
        )

    def percentage_complete(self):
        percent_value = self.percent_complete if self.percent_complete else 0

        return self.calculate_percentage(percent_value)

    @staticmethod
    def calculate_percentage(value):
        return round(value * 100)


class OpportunityStage(TimeStampedModel):
    name = models.CharField(max_length=50)

    class Meta:
        ordering = ('name', )

    def __str__(self):
        return self.name


class AvailableOpportunityStatusManager(models.Manager):
    """
    Return only Opportunity Statuses whose inactive field is False.
    """
    def get_queryset(self):
        return super().get_queryset().filter(
            inactive_flag=False
        )


class OpportunityStatus(TimeStampedModel):
    name = models.CharField(max_length=30)
    won_flag = models.BooleanField(default=False)
    lost_flag = models.BooleanField(default=False)
    closed_flag = models.BooleanField(default=False)
    inactive_flag = models.BooleanField(default=False)

    objects = models.Manager()
    available_objects = AvailableOpportunityStatusManager()

    class Meta:
        ordering = ('name', )
        verbose_name_plural = 'Opportunity statuses'

    def __str__(self):
        return self.name


class OpportunityPriority(TimeStampedModel):
    name = models.CharField(max_length=50)

    class Meta:
        ordering = ('name', )
        verbose_name_plural = 'opportunity priorities'

    def __str__(self):
        return self.name


class OpportunityType(TimeStampedModel):
    description = models.CharField(max_length=50)
    inactive_flag = models.BooleanField(default=False)

    class Meta:
        ordering = ('description', )

    def __str__(self):
        return self.description


class Opportunity(UpdateConnectWiseMixin, TimeStampedModel):
    business_unit_id = models.IntegerField(null=True)
    closed_date = models.DateTimeField(blank=True, null=True)
    customer_po = models.CharField(max_length=100, blank=True, null=True)
    date_became_lead = models.DateTimeField(blank=True, null=True)
    expected_close_date = models.DateField()
    location_id = models.IntegerField()
    name = models.CharField(max_length=100, db_index=True)
    notes = models.TextField(blank=True, null=True)
    pipeline_change_date = models.DateTimeField(blank=True, null=True)
    probability = models.ForeignKey('SalesProbability',
                                    blank=True, null=True,
                                    related_name='sales_probability',
                                    on_delete=models.SET_NULL)
    source = models.CharField(max_length=100, blank=True, null=True)

    closed_by = models.ForeignKey('Member',
                                  blank=True, null=True,
                                  related_name='opportunity_closed_by',
                                  on_delete=models.SET_NULL)
    company = models.ForeignKey('Company', blank=True, null=True,
                                related_name='company_opportunities',
                                on_delete=models.SET_NULL)
    contact = models.ForeignKey('Contact', blank=True, null=True,
                                on_delete=models.SET_NULL)
    primary_sales_rep = models.ForeignKey('Member',
                                          blank=True, null=True,
                                          related_name='opportunity_primary',
                                          on_delete=models.SET_NULL)
    priority = models.ForeignKey('OpportunityPriority',
                                 on_delete=models.SET_NULL, null=True)
    stage = models.ForeignKey('OpportunityStage', blank=True, null=True,
                              on_delete=models.SET_NULL)
    status = models.ForeignKey('OpportunityStatus', blank=True, null=True,
                               on_delete=models.SET_NULL)
    secondary_sales_rep = models.ForeignKey(
        'Member',
        blank=True, null=True,
        related_name='opportunity_secondary',
        on_delete=models.SET_NULL)
    opportunity_type = models.ForeignKey('OpportunityType',
                                         blank=True, null=True,
                                         on_delete=models.SET_NULL)
    udf = models.JSONField(blank=True, null=True)

    class Meta:
        ordering = ('name', )
        verbose_name_plural = 'Opportunities'

    def __str__(self):
        return self.name

    @property
    def api_class(self):
        return api.SalesAPIClient

    def _update_cw(self, api_client, updated_objects):
        return api_client.update_opportunity(self, updated_objects)

    def get_descriptor(self):
        return self.name

    def get_connectwise_url(self):
        params = dict(
            recordType='OpportunityFv',
            recid=self.id,
            companyName=settings.CONNECTWISE_CREDENTIALS['company_id']
        )
        return '{}/{}?{}'.format(
            settings.CONNECTWISE_SERVER_URL,
            settings.CONNECTWISE_TICKET_PATH,
            urllib.parse.urlencode(params)
        )


class Ticket(UpdateConnectWiseMixin, TimeStampedModel):
    SCHEDULE_ENTRY_TYPE = "S"

    PROJECT_TICKET = 'ProjectTicket'
    PROJECT_ISSUE = 'ProjectIssue'
    SERVICE_TICKET = 'ServiceTicket'

    RECORD_TYPES = (
        (SERVICE_TICKET, "Service Ticket"),
        (PROJECT_TICKET, "Project Ticket"),
        (PROJECT_ISSUE, "Project Issue"),
    )

    BILL_TIME_TYPES = (
        ('Billable', "Billable"),
        ('DoNotBill', "Do Not Bill"),
        ('NoCharge', "No Charge"),
        ('NoDefault', "No Default")
    )

    RESPOND = 'respond'
    PLAN = 'plan'
    RESOLVE = 'resolve'
    RESOLVED = 'resolved'
    WAITING = 'waiting'

    SLA_STAGE = (
        (RESPOND, "Respond"),
        (PLAN, "Plan"),
        (RESOLVE, "Resolve"),
        (RESOLVED, "Resolved"),
        (WAITING, "Waiting"),
    )

    STAGE_RANK = dict(
        zip(
            [i for i in range(5)],
            [type[0] for type in SLA_STAGE]
        )
    )
    TICKET = 'Ticket'
    PHASE = 'Phase'
    PREDECESSOR_TYPES = (
        (TICKET, 'Ticket'),
        (PHASE, 'Phase')
    )

    actual_hours = models.DecimalField(
        blank=True, null=True, decimal_places=2, max_digits=9)
    approved = models.BooleanField(null=True)
    budget_hours = models.DecimalField(
        blank=True, null=True, decimal_places=2, max_digits=12)
    closed_by = models.CharField(blank=True, null=True, max_length=250)
    closed_date_utc = models.DateTimeField(blank=True, null=True)
    closed_flag = models.BooleanField(null=True)
    customer_updated = models.BooleanField(default=False)
    date_resolved_utc = models.DateTimeField(blank=True, null=True)
    date_resplan_utc = models.DateTimeField(blank=True, null=True)
    date_responded_utc = models.DateTimeField(blank=True, null=True)
    entered_date_utc = models.DateTimeField(blank=True, null=True)
    has_child_ticket = models.BooleanField(null=True)
    impact = models.CharField(blank=True, null=True, max_length=250)
    is_in_sla = models.BooleanField(null=True)
    last_updated_utc = models.DateTimeField(blank=True, null=True)
    parent_ticket_id = models.IntegerField(blank=True, null=True)
    record_type = models.CharField(blank=True, null=True,
                                   max_length=250, choices=RECORD_TYPES,
                                   db_index=True)
    required_date_utc = models.DateTimeField(blank=True, null=True)
    respond_mins = models.IntegerField(blank=True, null=True)
    resolve_mins = models.IntegerField(blank=True, null=True)
    resources = models.CharField(blank=True, null=True, max_length=250)
    res_plan_mins = models.IntegerField(blank=True, null=True)
    severity = models.CharField(blank=True, null=True, max_length=250)
    site_name = models.CharField(blank=True, null=True, max_length=250)
    summary = models.CharField(blank=True, null=True, db_index=True,
                               max_length=250)
    updated_by = models.CharField(blank=True, null=True, max_length=250)
    sla_expire_date = models.DateTimeField(blank=True, null=True)
    do_not_escalate_date = models.DateTimeField(blank=True, null=True)
    sla_stage = models.CharField(blank=True, null=True,
                                 max_length=250, choices=SLA_STAGE,
                                 db_index=True)
    minutes_waiting = models.PositiveIntegerField(default=0)
    bill_time = models.CharField(blank=True, null=True,
                                 max_length=20, choices=BILL_TIME_TYPES)
    automatic_email_cc_flag = models.BooleanField(default=False)
    automatic_email_cc = models.CharField(blank=True,
                                          null=True,
                                          max_length=1000)
    automatic_email_contact_flag = models.BooleanField(default=False)
    automatic_email_resource_flag = models.BooleanField(default=False)

    predecessor_type = models.CharField(blank=True, null=True, max_length=10,
                                        choices=PREDECESSOR_TYPES)
    predecessor_closed_flag = models.BooleanField(default=False)
    lag_days = models.IntegerField(blank=True, null=True)
    lag_non_working_days_flag = models.BooleanField(default=False)
    estimated_start_date = models.DateTimeField(blank=True, null=True)
    wbs_code = models.CharField(blank=True, null=True, max_length=50)
    udf = models.JSONField(blank=True, null=True)
    tasks_completed = models.PositiveSmallIntegerField(blank=True, null=True)
    tasks_total = models.PositiveSmallIntegerField(blank=True, null=True)
    is_issue_flag = models.BooleanField(default=False)
    contact_name = models.CharField(blank=True, null=True, max_length=62)
    contact_phone_number = models.CharField(blank=True,
                                            null=True, max_length=20)
    contact_phone_extension = models.CharField(blank=True,
                                               null=True, max_length=15)
    contact_email_address = models.CharField(blank=True,
                                             null=True, max_length=250)

    # Only used for creation, not synced.
    initial_description = models.CharField(
        blank=True, null=True, max_length=5000)

    ticket_predecessor = models.ForeignKey(
        'self', blank=True, null=True, on_delete=models.SET_NULL
    )
    phase_predecessor = models.ForeignKey(
        'ProjectPhase', blank=True, null=True, on_delete=models.SET_NULL
    )
    board = models.ForeignKey(
        'ConnectwiseBoard', blank=True, null=True, on_delete=models.SET_NULL)
    company = models.ForeignKey(
        'Company', blank=True, null=True, related_name='company_tickets',
        on_delete=models.SET_NULL)
    contact = models.ForeignKey('Contact', blank=True, null=True,
                                on_delete=models.SET_NULL)
    location = models.ForeignKey(
        'Location', blank=True, null=True, related_name='location_tickets',
        on_delete=models.SET_NULL)
    members = models.ManyToManyField(
        'Member', through='ScheduleEntry',
        related_name='member_tickets')
    owner = models.ForeignKey(
        'Member', blank=True, null=True, on_delete=models.SET_NULL)
    priority = models.ForeignKey(
        'TicketPriority', blank=True, null=True, on_delete=models.SET_NULL)
    project = models.ForeignKey(
        'Project', blank=True, null=True, related_name='project_tickets',
        on_delete=models.SET_NULL)
    phase = models.ForeignKey(
        'ProjectPhase', blank=True, null=True, related_name='phase_tickets',
        on_delete=models.SET_NULL)
    status = models.ForeignKey(
        'BoardStatus', blank=True, null=True, related_name='status_tickets',
        on_delete=models.SET_NULL)
    team = models.ForeignKey(
        'Team', blank=True, null=True, related_name='team_tickets',
        on_delete=models.SET_NULL)
    sla = models.ForeignKey(
        'Sla', blank=True, null=True, related_name='tickets',
        on_delete=models.SET_NULL)
    type = models.ForeignKey(
        'Type', blank=True, null=True, related_name='type_tickets',
        on_delete=models.SET_NULL)
    sub_type = models.ForeignKey(
        'Subtype', blank=True, null=True, related_name='subtype_tickets',
        on_delete=models.SET_NULL)
    sub_type_item = models.ForeignKey(
        'Item', blank=True, null=True, related_name='item_tickets',
        on_delete=models.SET_NULL)
    agreement = models.ForeignKey(
        'Agreement', blank=True, null=True, related_name='agreement_tickets',
        on_delete=models.SET_NULL)
    source = models.ForeignKey(
        'Source', blank=True, null=True, related_name='source_tickets',
        on_delete=models.SET_NULL)

    class Meta:
        verbose_name = 'Ticket'
        verbose_name_plural = 'Tickets'
        ordering = ('summary', )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.state_manager = StateMachineManager
        state_class = self.state_manager.get(self.sla_stage)
        self.sla_state = state_class(self) if state_class else None

    def __str__(self):
        return '{}-{}'.format(self.id, self.summary)

    @property
    def api_class(self):
        if self.record_type in [self.PROJECT_TICKET, self.PROJECT_ISSUE]:
            api_class = api.ProjectAPIClient
        else:
            api_class = api.ServiceAPIClient

        return api_class

    def _update_cw(self, api_client, updated_objects):
        # TODO
        #  This _update_cw method just passes in the field names and the
        #  new data to be updated from those field names of only the ones to be
        #  changed.
        #  Other than that, none of this needs to be on the model class, it
        #  is only here for backwards compatibility until issue 2861

        # TODO Right now updating and creating records is split between
        #  Synchronizers and models. We need to refactor to have all
        #  interaction with the PSA to be through synchronizers.
        return api_client.update_ticket(self, updated_objects)

    def get_descriptor(self):
        return self.summary

    def time_remaining(self):
        time_remaining = self.budget_hours
        if self.budget_hours and self.actual_hours:
            time_remaining = self.budget_hours - self.actual_hours
        return time_remaining

    def get_connectwise_url(self):
        params = dict(
            locale='en_US',
            recordType='ServiceFv',
            recid=self.id,
            companyName=settings.CONNECTWISE_CREDENTIALS['company_id']
        )
        return '{}/{}?{}'.format(
            settings.CONNECTWISE_SERVER_URL,
            settings.CONNECTWISE_TICKET_PATH,
            urllib.parse.urlencode(params)
        )

    def save(self, *args, **kwargs):
        self._warn_invalid_status()
        super().save(**kwargs)

    def _warn_invalid_status(self):
        """
        Warn if the status doesn't belong to the board. It seems that
        ConnectWise isn't particularly strict about enforcing that a ticket's
        status is valid for the ticket's board, so we won't enforce this.

        If status or board are None, then don't bother, since this can happen
        during sync jobs and it would be a lot of work to enforce at all the
        right times.
        """
        if self.status and self.board and self.status.board != self.board:
            logger.warning(
                "For ticket {}, {} (ID {}) is not a valid status for the "
                "ticket's ConnectWise board ({}, ID {}).".
                format(
                    self.id,
                    self.status.name,
                    self.status.id,
                    self.board,
                    self.board.id,
                )
            )

    def close(self, *args, **kwargs):
        """
        Set the ticket to a closed status for the board.
        """
        logger.info('Closing ticket %s' % self.id)
        closed_status = self.board.get_closed_status()
        if closed_status is None:
            raise InvalidStatusError(
                "There are no closed statuses on this ticket's ConnectWise "
                "board ({}). Its status has not been changed.".format(
                    self.board
                )
            )

        self.status = closed_status
        self.closed_flag = True

        kwargs['changed_fields'] = ['status', 'closed_flag']

        return self.save(*args, **kwargs)

    def calculate_sla_expiry(self, priority_change=None):
        # We can't guarantee that entered_date_utc won't be null.
        # In the case that it is null, reset the date fields to prevent
        # incorrect SLA calculations.
        if not self.entered_date_utc:
            self.sla_expire_date = None
            self.do_not_escalate_date = None
            self.sla_stage = None
            self.date_resplan_utc = None
            self.date_responded_utc = None
            return

        if not self.status:
            logger.error(
                'Ticket {}-{}, does not have a status set. '
                'Skipping SLA calculation.'.format(self.id, self.summary)
            )
            return

        if not self.sla:
            return

        # SLAP might exist, which may alter the SLA target time
        sla_priority = SlaPriority.objects.filter(
                                                  sla=self.sla,
                                                  priority=self.priority
                                                  ).first()
        if sla_priority:
            sla = sla_priority
        else:
            sla = self.sla

        new_stage = self.STAGE_RANK.get(
            self.status.get_status_rank()
        )

        calendar = self.sla.get_calendar(self.company)

        if not calendar:
            logger.info(
                'No calendar found for SLA {} on ticket {}'.format(
                    sla.id, self.id
                )
            )
            return

        if priority_change:
            # When only the SLA priority has changed, we don't need to
            # change the state, just re-calculate the next SLA expiry date.
            sla_hours = sla.get_stage_hours(new_stage)
            if sla_hours:
                calendar.next_phase_expiry(sla_hours, self)
        else:
            self.change_sla_state(new_stage, calendar, sla)

    def change_sla_state(self, new_stage, calendar, sla):
        valid_stage = self.sla_state.get_valid_next_state(new_stage) \
            if self.sla_state else new_stage

        if valid_stage:
            new_state = self.state_manager.get(valid_stage)(self)

            self.sla_state.leave(calendar, sla) if self.sla_state else None
            self.sla_state = new_state
            self.sla_state.enter(valid_stage, calendar, sla)


class ServiceNote(UpdateRecordMixin, TimeStampedModel):

    created_by = models.TextField(blank=True, null=True, max_length=250)
    date_created = models.DateTimeField(blank=True, null=True)
    detail_description_flag = models.BooleanField(blank=True)
    external_flag = models.BooleanField(blank=True)
    internal_analysis_flag = models.BooleanField(blank=True)
    internal_flag = models.BooleanField(blank=True)
    resolution_flag = models.BooleanField(blank=True)
    text = models.TextField(blank=True, null=True, max_length=4000)

    ticket = models.ForeignKey('Ticket', on_delete=models.CASCADE)
    member = models.ForeignKey(
        'Member', blank=True, null=True, on_delete=models.SET_NULL)

    class Meta:
        ordering = ('-date_created', 'id')
        verbose_name_plural = 'Notes'

    def __str__(self):
        return 'Ticket {} note: {}'.format(self.ticket, str(self.date_created))

    def get_entered_time(self):
        return self.date_created

    @property
    def note(self):
        return self.text

    def update_cw(self, **kwargs):
        api_class = self.ticket.api_class

        api_client = api_class(
            api_public_key=kwargs.get('api_public_key'),
            api_private_key=kwargs.get('api_private_key')
        )
        return api_client.update_note(self)


class Sla(TimeStampedModel, SlaGoalsMixin):

    BASED_ON = (
        ('MyCalendar', "My Company Calendar"),
        ('Customer', "Customer's Calendar"),
        ('AllHours', "24 Hours"),
        ('Custom', "Custom Calendar")
    )

    name = models.TextField(max_length=250)
    default_flag = models.BooleanField(default=False)
    respond_hours = models.BigIntegerField()
    plan_within = models.BigIntegerField()
    resolution_hours = models.BigIntegerField()
    based_on = models.CharField(max_length=50, choices=BASED_ON,
                                default='MyCalendar')

    calendar = models.ForeignKey(
        'Calendar',
        blank=True,
        null=True,
        on_delete=models.SET_NULL
        )

    class Meta:
        verbose_name_plural = 'SLAs'

    def __str__(self):
        return self.name

    def get_calendar(self, company):
        try:
            if self.calendar:
                return self.calendar
            elif self.based_on == 'Customer' and company:
                return Company.objects.get(id=company.id).calendar
            elif self.based_on == 'MyCalendar':
                # Using get instead of first so it will throw an exception
                return MyCompanyOther.objects.get().default_calendar
            else:
                # Maybe based_on was Customer but company was None
                return None
        except ObjectDoesNotExist:
            return None


class SlaPriority(TimeStampedModel, SlaGoalsMixin):

    sla = models.ForeignKey(
        'Sla',
        on_delete=models.CASCADE
        )
    priority = models.ForeignKey(
        'TicketPriority',
        on_delete=models.CASCADE
        )
    respond_hours = models.FloatField()
    plan_within = models.FloatField()
    resolution_hours = models.FloatField()

    class Meta:
        verbose_name_plural = 'SLA Priorities'

    def __str__(self):
        return 'priority: {}, on SLA:{}'.format(
            str(self.priority), self.sla.id)


class OpportunityNote(TimeStampedModel):

    text = models.TextField(blank=True, null=True, max_length=2000)
    date_created = models.DateTimeField(blank=True, null=True)
    opportunity = models.ForeignKey('Opportunity', on_delete=models.CASCADE)

    class Meta:
        ordering = ('-date_created', 'id', )
        verbose_name_plural = 'Opportunity notes'

    def __str__(self):
        return 'id: {}, on Opportunity id:{}'.format(
            str(self.id), self.opportunity.id)


class ActivityStatus(TimeStampedModel):
    name = models.CharField(max_length=50)
    default_flag = models.BooleanField(default=False)
    inactive_flag = models.BooleanField(default=False)
    spawn_followup_flag = models.BooleanField(default=False)
    closed_flag = models.BooleanField(default=False)

    class Meta:
        ordering = ('name', )
        verbose_name_plural = 'activity statuses'

    def __str__(self):
        return self.name


class ActivityType(TimeStampedModel):
    name = models.CharField(max_length=50)
    points = models.IntegerField()
    default_flag = models.BooleanField(default=False)
    inactive_flag = models.BooleanField(default=False)
    email_flag = models.BooleanField(default=False)
    memo_flag = models.BooleanField(default=False)
    history_flag = models.BooleanField(default=False)

    class Meta:
        ordering = ('name', )

    def __str__(self):
        return self.name


class Activity(UpdateConnectWiseMixin, TimeStampedModel):
    name = models.CharField(max_length=250)
    notes = models.TextField(blank=True, null=True, max_length=2000)
    date_start = models.DateTimeField(blank=True, null=True)
    date_end = models.DateTimeField(blank=True, null=True)

    assign_to = models.ForeignKey('Member', on_delete=models.CASCADE)
    opportunity = models.ForeignKey(
        'Opportunity', blank=True, null=True, on_delete=models.CASCADE)
    ticket = models.ForeignKey(
        'Ticket', blank=True, null=True, on_delete=models.CASCADE)
    status = models.ForeignKey(
        'ActivityStatus', blank=True, null=True, on_delete=models.SET_NULL)
    type = models.ForeignKey(
        'ActivityType', blank=True, null=True, on_delete=models.SET_NULL)
    company = models.ForeignKey(
        'Company', blank=True, null=True, on_delete=models.SET_NULL)
    contact = models.ForeignKey('Contact', blank=True, null=True,
                                on_delete=models.SET_NULL)
    agreement = models.ForeignKey(
        'Agreement', blank=True, null=True, on_delete=models.SET_NULL)
    udf = models.JSONField(blank=True, null=True)
    members = models.ManyToManyField(
        'Member',
        through='ScheduleEntry',
        related_name='member_activities'
    )

    class Meta:
        verbose_name_plural = 'activities'

    def __str__(self):
        return self.name or ''

    @property
    def api_class(self):
        return api.SalesAPIClient

    def _update_cw(self, api_client, updated_objects):
        return api_client.update_activity(self, updated_objects)

    def get_connectwise_url(self):
        params = dict(
            recordType='ActivityFv',
            recid=self.id,
            companyName=settings.CONNECTWISE_CREDENTIALS['company_id']
        )
        return '{}/{}?{}'.format(
            settings.CONNECTWISE_SERVER_URL,
            settings.CONNECTWISE_TICKET_PATH,
            urllib.parse.urlencode(params)
        )


class SalesProbability(TimeStampedModel):
    probability = models.IntegerField()

    class Meta:
        verbose_name_plural = 'Sales probabilities'
        ordering = ('probability', )

    def __str__(self):
        return 'Probability {}'.format(self.probability)


class SLAMachineState(object):
    valid_next_states = set()

    def __init__(self, ticket):
        self.ticket = ticket

    def enter(self, new_stage, calendar, sla):
        logger.debug('Entering SLA State: {}'.format(self))
        sla_hours = sla.get_stage_hours(new_stage)
        calendar.next_phase_expiry(sla_hours, self.ticket)
        self.ticket.sla_stage = new_stage

    def leave(self, *args):
        logger.debug('Leaving SLA State: {}'.format(self))

    def get_valid_next_state(self, new_state):
        if new_state in self.valid_next_states:
            return new_state


class EscalateState(SLAMachineState):
    valid_next_states = {Ticket.WAITING}


class Respond(EscalateState):
    sla_stage = Ticket.RESPOND

    def __init__(self, ticket):
        super().__init__(ticket)
        self.valid_next_states = self.valid_next_states.union({
                                  Ticket.PLAN,
                                  Ticket.RESOLVE,
                                  Ticket.RESOLVED
                             })


class Plan(EscalateState):
    sla_stage = Ticket.PLAN

    def __init__(self, ticket):
        super().__init__(ticket)
        self.valid_next_states = self.valid_next_states.union({
                                  Ticket.RESOLVE,
                                  Ticket.RESOLVED
                             })


class Resolve(EscalateState):
    sla_stage = Ticket.RESOLVE

    def __init__(self, ticket):
        super().__init__(ticket)
        self.valid_next_states = \
            self.valid_next_states.union({Ticket.RESOLVED})


class Resolved(EscalateState):
    sla_stage = Ticket.RESOLVED

    def __init__(self, ticket):
        super().__init__(ticket)
        self.valid_next_states = self.valid_next_states.union({Ticket.RESOLVE})

    def enter(self, *args):
        self.ticket.sla_expire_date = None
        self.ticket.sla_stage = self.sla_stage

    def get_valid_next_state(self, new_state):
        if new_state == Ticket.WAITING:
            return new_state
        elif new_state != self.sla_stage:
            return Ticket.RESOLVE


class Waiting(SLAMachineState):
    sla_stage = Ticket.WAITING

    def enter(self, *args):
        # Save time entered a non-escalate state to calculate how many minutes
        # it has been waiting if it is ever returned to a regular state
        self.ticket.sla_expire_date = None
        self.ticket.do_not_escalate_date = timezone.now()
        self.ticket.sla_stage = self.sla_stage

    def leave(self, calendar, sla):
        # Get the minutes ticket has been in a non-escalate state,
        # change the state to the desired stage, or the lowest stage allowed,
        # add the minutes spent in waiting to the tickets minutes_waiting field
        # and then recalculate the new stage
        if self.ticket.do_not_escalate_date:
            self.ticket.minutes_waiting += calendar.get_sla_time(
                self.ticket.do_not_escalate_date.astimezone(tz=None),
                timezone.now().astimezone(tz=None)
                )
        self.ticket.do_not_escalate_date = None

    def get_valid_next_state(self, new_state):
        if new_state != Ticket.WAITING:
            return self._lowest_possible_stage(new_state)

    def _lowest_possible_stage(self, stage):
        # Returns the lowest stage a ticket is allowed to go, given the input
        # stage.
        if stage == Ticket.RESOLVED or stage == Ticket.RESOLVE:
            return stage
        elif stage == Ticket.PLAN:
            if self.ticket.date_resplan_utc:
                return Ticket.RESOLVE
            else:
                return stage
        elif stage == Ticket.RESPOND:
            if self.ticket.date_resplan_utc:
                return Ticket.RESOLVE
            elif self.ticket.date_responded_utc:
                return Ticket.PLAN
            else:
                return stage
        else:
            logger.warning('Exiting stage with unknown type')
            return stage


class StateMachineManager(object):
    SLA_STATE = {
        Ticket.RESPOND: Respond,
        Ticket.PLAN: Plan,
        Ticket.RESOLVE: Resolve,
        Ticket.RESOLVED: Resolved,
        Ticket.WAITING: Waiting,
    }

    @classmethod
    def get(cls, state):
        return cls.SLA_STATE.get(state)


class Type(TimeStampedModel):
    name = models.CharField(max_length=50)
    board = models.ForeignKey('ConnectWiseBoard', on_delete=models.CASCADE)
    inactive_flag = models.BooleanField(default=False)

    class Meta:
        verbose_name = 'Type'
        verbose_name_plural = 'Types'

    def __str__(self):
        return "{} - {}".format(self.name, self.board)


class SubType(TimeStampedModel):
    name = models.CharField(max_length=50)
    board = models.ForeignKey('ConnectWiseBoard', on_delete=models.CASCADE)
    inactive_flag = models.BooleanField(default=False)

    class Meta:
        verbose_name = 'Subtype'
        verbose_name_plural = 'Subtypes'

    def __str__(self):
        return self.name


class Item(TimeStampedModel):
    name = models.CharField(max_length=50)
    board = models.ForeignKey('ConnectWiseBoard', on_delete=models.CASCADE)
    inactive_flag = models.BooleanField(default=False)

    class Meta:
        verbose_name = 'Item'
        verbose_name_plural = 'Items'

    def __str__(self):
        return self.name


class TypeSubTypeItemAssociation(TimeStampedModel):
    type = models.ForeignKey('Type', blank=True, null=True,
                             related_name='associations',
                             on_delete=models.CASCADE)
    sub_type = models.ForeignKey('SubType', blank=True,
                                 related_name='associations',
                                 null=True, on_delete=models.CASCADE)
    item = models.ForeignKey('Item', blank=True,
                             related_name='associations',
                             null=True, on_delete=models.CASCADE)
    board = models.ForeignKey('ConnectWiseBoard', blank=True,
                              null=True, on_delete=models.CASCADE)


class WorkType(TimeStampedModel):
    name = models.CharField(max_length=50)
    inactive_flag = models.BooleanField(default=False)
    overall_default_flag = models.BooleanField(default=False)
    bill_time = models.CharField(
        max_length=50, choices=BILL_TYPES, blank=True, null=True
    )

    def __str__(self):
        return self.name


class WorkRole(TimeStampedModel):
    name = models.CharField(max_length=50)
    inactive_flag = models.BooleanField(default=False)

    def __str__(self):
        return self.name


class Agreement(TimeStampedModel):
    name = models.CharField(max_length=100)
    agreement_type = models.CharField(max_length=50, null=True)
    agreement_status = models.CharField(max_length=50, blank=True, null=True)
    cancelled_flag = models.BooleanField(default=False)
    bill_time = models.CharField(
        max_length=50, choices=BILL_TYPES, blank=True, null=True
    )
    work_type = models.ForeignKey(
        'WorkType', null=True, on_delete=models.SET_NULL)
    work_role = models.ForeignKey(
        'WorkRole', null=True, on_delete=models.SET_NULL)
    company = models.ForeignKey(
        'Company', null=True, on_delete=models.SET_NULL)

    def __str__(self):
        return '{}/{}'.format(self.agreement_type, self.name)


class Source(TimeStampedModel):
    name = models.CharField(max_length=100, blank=False, null=False)
    default_flag = models.BooleanField()

    def __str__(self):
        return self.name


class BaseUDF(TimeStampedModel):
    caption = models.CharField(max_length=50, blank=True, null=True)
    type = models.CharField(max_length=50, blank=True, null=True)
    entry_method = models.CharField(max_length=50, blank=True, null=True)
    number_of_decimals = \
        models.PositiveSmallIntegerField(blank=True, null=True)

    class Meta:
        abstract = True

    def __str__(self):
        return self.caption


class TicketUDF(BaseUDF):
    pass


class ProjectUDF(BaseUDF):
    pass


class ActivityUDF(BaseUDF):
    pass


class OpportunityUDF(BaseUDF):
    pass


class TicketTracker(Ticket):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_ticket'


class ConnectWiseBoardTracker(ConnectWiseBoard):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_connectwiseboard'


class BoardStatusTracker(BoardStatus):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_boardstatus'


class LocationTracker(Location):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_location'


class MemberTracker(Member):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_member'


class CompanyTracker(Company):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_company'


class CompanyStatusTracker(CompanyStatus):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_companystatus'


class CompanyTypeTracker(CompanyType):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_companytype'


class CompanyNoteTypeTracker(CompanyNoteType):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_companynotetype'


class CompanyTeamTracker(CompanyTeam):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_companyteam'


class CompanyTeamRoleTracker(CompanyTeamRole):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_companyteamrole'


class ContactTracker(Contact):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_contact'


class ContactCommunicationTracker(ContactCommunication):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_contactcommunication'


class CommunicationTypeTracker(CommunicationType):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_communicationtype'


class MyCompanyOtherTracker(MyCompanyOther):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_mycompanyother'


class CalendarTracker(Calendar):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_calendar'


class HolidayTracker(Holiday):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_holiday'


class HolidayListTracker(HolidayList):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_holidaylist'


class ScheduleTypeTracker(ScheduleType):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_scheduletype'


class ScheduleStatusTracker(ScheduleStatus):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_schedulestatus'


class ScheduleEntryTracker(ScheduleEntry):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_scheduleentry'


class TerritoryTracker(Territory):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_territory'


class TimeEntryTracker(TimeEntry):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_timeentry'


class TeamTracker(Team):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_team'


class TicketPriorityTracker(TicketPriority):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_ticketpriority'


class ProjectStatusTracker(ProjectStatus):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_projectstatus'


class ProjectTypeTracker(ProjectType):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_projecttype'


class ProjectPhaseTracker(ProjectPhase):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_projectphase'


class ProjectTeamMemberTracker(ProjectTeamMember):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_projectteammember'


class ProjectTracker(Project):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_project'


class OpportunityStageTracker(OpportunityStage):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_opportunitystage'


class OpportunityStatusTracker(OpportunityStatus):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_opportunitystatus'


class OpportunityPriorityTracker(OpportunityPriority):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_opportunitypriority'


class OpportunityTypeTracker(OpportunityType):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_opportunitytype'


class OpportunityTracker(Opportunity):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_opportunity'


class ServiceNoteTracker(ServiceNote):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_servicenote'


class SlaTracker(Sla):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_sla'


class SlaPriorityTracker(SlaPriority):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_slapriority'


class OpportunityNoteTracker(OpportunityNote):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_opportunitynote'


class ActivityStatusTracker(ActivityStatus):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_activitystatus'


class ActivityTypeTracker(ActivityType):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_activitytype'


class ActivityTracker(Activity):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_activity'


class SalesProbabilityTracker(SalesProbability):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_salesprobability'


class SourceTracker(Source):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_source'


class TypeTracker(Type):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_type'


class SubTypeTracker(SubType):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_subtype'


class ItemTracker(Item):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_item'


class TypeSubTypeItemAssociationTracker(TypeSubTypeItemAssociation):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_type_subtype_item_association'


class WorkTypeTracker(WorkType):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_worktype'


class WorkRoleTracker(WorkRole):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_workrole'


class AgreementTracker(Agreement):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_agreement'


class TicketUDFTracker(TicketUDF):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_ticketudf'


class ProjectUDFTracker(ProjectUDF):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_projectudf'


class ActivityUDFTracker(ActivityUDF):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_activityudf'


class OpportunityUDFTracker(OpportunityUDF):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djconnectwise_opportunityudf'



# Generated by Django 3.1.7 on 2021-07-23 10:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0152_auto_20210722_1224'),
    ]

    operations = [
        migrations.AlterField(
            model_name='contact',
            name='first_name',
            field=models.CharField(blank=True, db_index=True, max_length=200, null=True),
        ),
        migrations.AlterField(
            model_name='contact',
            name='last_name',
            field=models.CharField(blank=True, db_index=True, max_length=200, null=True),
        ),
    ]



# Generated by Django 4.0.4 on 2022-06-22 16:30

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0163_ticket_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='ticket',
            name='source',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='source_tickets', to='djconnectwise.source'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0018_auto_20170505_1531'),
    ]

    operations = [
        migrations.AddField(
            model_name='syncjob',
            name='entity_name',
            field=models.CharField(max_length=100, default=''),
            preserve_default=False,
        ),
    ]



# Generated by Django 2.1.14 on 2020-02-04 10:54

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0112_syncjob_synchronizer_class'),
    ]

    operations = [
        migrations.AlterField(
            model_name='opportunity',
            name='stage',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='djconnectwise.OpportunityStage'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django_extensions.db.fields


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0048_project_manager'),
    ]

    operations = [
        migrations.CreateModel(
            name='SalesProbability',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('probability', models.IntegerField()),
            ],
            options={
                'ordering': ('probability',),
                'verbose_name_plural': 'Sales probabilities',
            },
        ),
        migrations.AddField(
            model_name='opportunity',
            name='probability',
            field=models.ForeignKey(null=True, to='djconnectwise.SalesProbability', blank=True, related_name='sales_probability', on_delete=models.CASCADE),
        ),
    ]



# Generated by Django 3.1.2 on 2020-10-16 08:10

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0134_auto_20201015_1804'),
    ]

    operations = [
        migrations.AddField(
            model_name='activity',
            name='udf',
            field=models.JSONField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='opportunity',
            name='udf',
            field=models.JSONField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='project',
            name='udf',
            field=models.JSONField(blank=True, null=True),
        ),
    ]



# Generated by Django 3.1.5 on 2021-06-10 17:41

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0149_auto_20210427_2236'),
    ]

    operations = [
        migrations.AddField(
            model_name='ticket',
            name='estimated_start_date',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='wbs_code',
            field=models.CharField(blank=True, max_length=50, null=True),
        ),
    ]



# Generated by Django 2.1.11 on 2020-04-16 09:33

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0114_auto_20200219_1515'),
    ]

    operations = [
        migrations.AddField(
            model_name='agreement',
            name='agreement_status',
            field=models.CharField(blank=True, max_length=50, null=True),
        ),
    ]
